import os


def local_scheme(version):
    return ""


def get_version():
    from setuptools_scm import get_version as default_version

    github_version = os.environ.get("GITHUB_REF_NAME", None)
    github_type = os.environ.get("GITHUB_REF_TYPE", None)
    if github_version is not None and github_type == "tag":
        print(f"GITHUB_REF_NAME found, using version: {github_version}")
        return github_version
    return default_version(local_scheme=local_scheme)



# coding: utf-8

"""Webmail tests."""

from __future__ import unicode_literals

import os
import shutil
import tempfile

try:
    import mock
except ImportError:
    from unittest import mock
from six import BytesIO

from django.core import mail
from django.urls import reverse

from kalabash.admin import factories as admin_factories
from kalabash.core import models as core_models
from kalabash.lib.tests import KbashTestCase

from . import data as tests_data


BODYSTRUCTURE_SAMPLE_WITH_FLAGS = [
    (b'19 (UID 19 FLAGS (\\Seen) RFC822.SIZE 100000 BODYSTRUCTURE (("text" "plain" ("charset" "ISO-8859-1" "format" "flowed") NIL NIL "7bit" 2 1 NIL NIL NIL NIL)("message" "rfc822" ("name*" "ISO-8859-1\'\'%5B%49%4E%53%43%52%49%50%54%49%4F%4E%5D%20%52%E9%63%E9%70%74%69%6F%6E%20%64%65%20%76%6F%74%72%65%20%64%6F%73%73%69%65%72%20%64%27%69%6E%73%63%72%69%70%74%69%6F%6E%20%46%72%65%65%20%48%61%75%74%20%44%E9%62%69%74") NIL NIL "8bit" 3632 ("Wed, 13 Dec 2006 20:30:02 +0100" {70}',  # noqa
     b"[INSCRIPTION] R\xe9c\xe9ption de votre dossier d'inscription Free Haut D\xe9bit"),  # noqa
    (b' (("Free Haut Debit" NIL "inscription" "freetelecom.fr")) (("Free Haut Debit" NIL "inscription" "freetelecom.fr")) ((NIL NIL "hautdebit" "freetelecom.fr")) ((NIL NIL "nguyen.antoine" "wanadoo.fr")) NIL NIL NIL "<20061213193125.9DA0919AC@dgroup2-2.proxad.net>") ("text" "plain" ("charset" "iso-8859-1") NIL NIL "8bit" 1428 38 NIL ("inline" NIL) NIL NIL) 76 NIL ("inline" ("filename*" "ISO-8859-1\'\'%5B%49%4E%53%43%52%49%50%54%49%4F%4E%5D%20%52%E9%63%E9%70%74%69%6F%6E%20%64%65%20%76%6F%74%72%65%20%64%6F%73%73%69%65%72%20%64%27%69%6E%73%63%72%69%70%74%69%6F%6E%20%46%72%65%65%20%48%61%75%74%20%44%E9%62%69%74")) NIL NIL) "mixed" ("boundary" "------------040706080908000209030901") NIL NIL NIL) BODY[HEADER.FIELDS (DATE FROM TO CC SUBJECT)] {266}',  # noqa
     b'Date: Tue, 19 Dec 2006 19:50:13 +0100\r\nFrom: Alphamonak Solutions <nguyen.antoine@wanadoo.fr>\r\nTo: Alphamonak Solutions <amonak@alphamonak.com>\r\nSubject: [Fwd: [INSCRIPTION] =?ISO-8859-1?Q?R=E9c=E9ption_de_votre_?=\r\n =?ISO-8859-1?Q?dossier_d=27inscription_Free_Haut_D=E9bit=5D?=\r\n\r\n'
    ),
    b')'
]


def get_gif():
    """Return gif."""
    gif = BytesIO(
        b"GIF87a\x01\x00\x01\x00\x80\x01\x00\x00\x00\x00ccc,\x00"
        b"\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;")
    gif.name = "image.gif"
    return gif


class IMAP4Mock(object):
    """Fake IMAP4 client."""

    def __init__(self, *args, **kwargs):
        self.untagged_responses = {}

    def _quote(self, data):
        return data

    def _simple_command(self, name, *args, **kwargs):
        if name == "CAPABILITY":
            self.untagged_responses["CAPABILITY"] = [b""]
        elif name == "LIST":
            self.untagged_responses["LIST"] = [b"() \".\" \"INBOX\""]
        elif name == "NAMESPACE":
            self.untagged_responses["NAMESPACE"] = [b'(("" "/")) NIL NIL']
        return "OK", None

    def append(self, *args, **kwargs):
        pass

    def create(self, name):
        return "OK", None

    def delete(self, name):
        return "OK", None

    def list(self):
        return "OK", [b"() \".\" \"INBOX\""]

    def rename(self, oldname, newname):
        return "OK", None

    def uid(self, command, *args):
        if command == "SORT":
            return "OK", [b"19"]
        elif command == "FETCH":
            uid = int(args[0])
            data = BODYSTRUCTURE_SAMPLE_WITH_FLAGS
            if uid == 46931:
                if args[1] == "(BODYSTRUCTURE)":
                    data = tests_data.BODYSTRUCTURE_ONLY_4
                elif "HEADER.FIELDS" in args[1]:
                    data = tests_data.BODYSTRUCTURE_SAMPLE_4
                else:
                    data = tests_data.BODY_PLAIN_4
            elif uid == 46932:
                if args[1] == "(BODYSTRUCTURE)":
                    data = tests_data.BODYSTRUCTURE_ONLY_5
                elif "HEADER.FIELDS" in args[1]:
                    data = tests_data.BODYSTRUCTURE_SAMPLE_9
                else:
                    data = tests_data.BODYSTRUCTURE_SAMPLE_10
            elif uid == 33:
                if args[1] == "(BODYSTRUCTURE)":
                    data = tests_data.BODYSTRUCTURE_EMPTY_MAIL
                else:
                    data = tests_data.EMPTY_BODY
            elif uid == 133872:
                data = tests_data.COMPLETE_MAIL
            return "OK", data
        elif command == "STORE":
            return "OK", []


class WebmailTestCase(KbashTestCase):
    """Check webmail backend."""

    @classmethod
    def setUpTestData(cls):  # noqa
        """Create some users."""
        super(WebmailTestCase, cls).setUpTestData()
        admin_factories.populate_database()
        cls.user = core_models.User.objects.get(username="user@test.com")

    def setUp(self):
        """Connect with a simpler user."""
        patcher = mock.patch("imaplib.IMAP4")
        self.mock_imap4 = patcher.start()
        self.mock_imap4.return_value = IMAP4Mock()
        self.addCleanup(patcher.stop)
        self.set_global_parameter("imap_port", 1435)
        self.workdir = tempfile.mkdtemp()
        os.mkdir("{}/webmail".format(self.workdir))
        self.set_global_parameter("update_scheme", False, app="core")
        url = reverse("core:login")
        data = {
            "username": self.user.username, "password": "toto"
        }
        self.client.post(url, data)

    def tearDown(self):
        """Cleanup."""
        shutil.rmtree(self.workdir)

    def test_listmailbox(self):
        """Check listmailbox action."""
        url = reverse("kalabash_webmail:index")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            "{}?action=listmailbox".format(url),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            "nguyen.antoine@wanadoo.fr", response.json()["listing"])

        response = self.client.get(
            "{}?action=listmailbox&pattern=Réception&criteria=Subject"
            .format(url),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertIn(
            "nguyen.antoine@wanadoo.fr", response.json()["listing"])

    def test_attachments(self):
        """Check attachments."""
        url = reverse("kalabash_webmail:index")
        response = self.client.get("{}?action=compose".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertIn("compose_mail", self.client.session)
        url = reverse("kalabash_webmail:attachment_list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        self.set_global_parameters({"max_attachment_size": "10"})
        with self.settings(MEDIA_ROOT=self.workdir):
            response = self.client.post(url, {"attachment": get_gif()})
        self.assertContains(response, "Attachment is too big")

        self.set_global_parameters({"max_attachment_size": "10K"})
        with self.settings(MEDIA_ROOT=self.workdir):
            response = self.client.post(url, {"attachment": get_gif()})
        self.assertContains(response, "upload_success")
        self.assertEqual(
            len(self.client.session["compose_mail"]["attachments"]), 1)
        name = self.client.session["compose_mail"]["attachments"][0]["tmpname"]
        path = "{}/webmail/{}".format(self.workdir, name)
        self.assertTrue(os.path.exists(path))

        url = reverse("kalabash_webmail:attachment_delete")
        with self.settings(MEDIA_ROOT=self.workdir):
            self.ajax_get("{}?name={}".format(url, name))
        self.assertFalse(os.path.exists(path))

    def test_delattachment_errors(self):
        """Check error cases."""
        url = reverse("kalabash_webmail:index")
        response = self.client.get("{}?action=compose".format(url))
        self.assertEqual(response.status_code, 200)
        self.assertIn("compose_mail", self.client.session)

        url = reverse("kalabash_webmail:attachment_delete")
        with self.settings(MEDIA_ROOT=self.workdir):
            response = self.ajax_get("{}?name=".format(url))
        self.assertEqual(response["status"], "ko")
        self.assertEqual(response["respmsg"], "Bad query")

        with self.settings(MEDIA_ROOT=self.workdir):
            response = self.ajax_get("{}?name=test".format(url))
        self.assertEqual(response["status"], "ko")
        self.assertEqual(response["respmsg"], "Unknown attachment")

    def test_send_mail(self):
        """Check compose form."""
        url = "{}?action=compose".format(reverse("kalabash_webmail:index"))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            url, {
                "from_": self.user.email, "to": "test@example.test",
                "subject": "test", "body": "Test"
            }
        )
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].from_email, "user@test.com")

        # Try to send an email using HTML format
        self.user.first_name = "Antoine"
        self.user.last_name = "Nguyen"
        self.user.parameters.set_value("editor", "html")
        self.user.save()
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        mail.outbox = []
        response = self.client.post(
            url, {
                "from_": self.user.email,
                "to": "test@example.test", "subject": "test",
                "body": "<p>Test</p>"
            }
        )
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].from_email, '"Alphamonak Solutions" <user@test.com>')

    def test_signature(self):
        """Check signature in different formats."""
        signature = "Alphamonak Solutions"
        self.user.parameters.set_value("signature", signature)
        self.user.save()

        response = self.client.get(reverse("kalabash_webmail:index"))
        self.assertEqual(response.status_code, 200)

        url = "{}?action=compose".format(reverse("kalabash_webmail:index"))
        response = self.ajax_get(url)
        self.assertIn(signature, response["listing"])

    def test_custom_js_in_preferences(self):
        """Check that custom js is included."""
        url = reverse("core:user_index")
        response = self.client.get(url)
        self.assertContains(response, "function toggleSignatureEditor()")

    def test_send_mail_errors(self):
        """Check error cases."""
        url = "{}?action=compose".format(reverse("kalabash_webmail:index"))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.ajax_post(
            url, {"to": "", "subject": "test", "body": "Test"}, 400
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_new_folder(self):
        """Test folder creation."""
        url = reverse("kalabash_webmail:folder_add")
        response = self.client.get(url)
        self.assertContains(response, "Create a new folder")

        response = self.ajax_post(url, {"name": "Test"})
        self.assertIn("newmb", response)

    def test_edit_folder(self):
        """Test folder edition."""
        url = reverse("kalabash_webmail:folder_change")
        response = self.client.get(url)
        self.assertContains(response, "Invalid request")

        url = "{}?name=Test".format(url)
        response = self.client.get(url)
        self.assertContains(response, "Edit folder")

        session = self.client.session
        session["webmail_navparams"] = {"inbox": "Test"}
        session.save()
        response = self.ajax_post(url, {"oldname": "Test", "name": "Toto"})
        self.assertEqual(response["respmsg"], "Folder updated")

    def test_delete_folder(self):
        """Test folder removal."""
        url = reverse("kalabash_webmail:folder_delete")
        self.ajax_get(url, status=400)

        url = "{}?name=Test".format(url)
        session = self.client.session
        session["webmail_navparams"] = {"inbox": "Test"}
        session.save()
        self.ajax_get(url)

    def test_reply_to_email(self):
        """Test reply form."""
        url = "{}?action=reply&mbox=INBOX&mailid=46931".format(
            reverse("kalabash_webmail:index"))
        session = self.client.session
        session["lastaction"] = "compose"
        session.save()
        response = self.ajax_get(url)
        self.assertIn('id="id_origmsgid"', response["listing"])

        response = self.client.post(
            url, {
                "from_": self.user.email, "to": "test@example.test",
                "subject": "test", "body": "Test",
                "origmsgid": "<id@localhost>"
            }
        )
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].from_email, "user@test.com")
        self.assertIn("References", mail.outbox[0].extra_headers)

    def test_forward_email(self):
        """Test forward form."""
        url = "{}?action=forward&mbox=INBOX&mailid=46932".format(
            reverse("kalabash_webmail:index"))
        session = self.client.session
        session["lastaction"] = "compose"
        session.save()
        with self.settings(MEDIA_ROOT=self.workdir):
            response = self.client.get(
                url, HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        response = response.json()
        self.assertIn('id="id_origmsgid"', response["listing"])
        self.assertEqual(
            len(self.client.session["compose_mail"]["attachments"]), 1)
        response = self.client.post(
            url, {
                "from_": self.user.email, "to": "test@example.test",
                "subject": "test", "body": "Test",
                "origmsgid": "<id@localhost>"
            }
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_getmailcontent_empty_mail(self):
        """Try to display an empty email."""
        url = "{}?action=reply&mbox=INBOX&mailid=33".format(
            reverse("kalabash_webmail:mailcontent_get"))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_getmailsource(self):
        """Try to display a message's source."""
        url = "{}?mbox=INBOX&mailid=133872".format(
            reverse("kalabash_webmail:mailsource_get"))
        response = self.client.get(url)
        self.assertContains(response, "Message-ID")



"""Custom form validators."""

from email.utils import getaddresses

from django.utils.encoding import force_str
from django.core.validators import validate_email


class EmailListValidator(object):

    """Validate a list of email."""

    def __call__(self, value):
        value = force_str(value)
        addresses = getaddresses([value])
        [validate_email(email) for name, email in addresses if email]


validate_email_list = EmailListValidator()



# coding: utf-8

"""FETCH parser tests."""

from __future__ import print_function

import unittest

import six

from kalabash_webmail.lib.fetch_parser import FetchResponseParser

from . import data


def dump_bodystructure(fp, bs, depth=0):
    """Dump a parsed BODYSTRUCTURE."""
    indentation = " " * (depth * 4)
    for mp in bs:
        if isinstance(mp, list):
            if isinstance(mp[0], list):
                print("{}multipart/{}".format(indentation, mp[1]), file=fp)
                dump_bodystructure(fp, mp, depth + 1)
            else:
                dump_bodystructure(fp, mp, depth)
        elif isinstance(mp, dict):
            if isinstance(mp["struct"][0], list):
                print("{}multipart/{}".format(
                    indentation, mp["struct"][1]), file=fp)
                dump_bodystructure(fp, mp["struct"][0], depth + 1)
            else:
                print("{}{}/{}".format(
                    indentation, *mp["struct"][:2]), file=fp)
    fp.seek(0)
    result = fp.read()
    return result


class FetchParserTestCase(unittest.TestCase):
    """Test FETCH parser."""

    def setUp(self):
        """Setup test env."""
        self.parser = FetchResponseParser()

    def _test_bodystructure_output(self, bs, expected):
        """."""
        r = self.parser.parse(bs)
        fp = six.StringIO()
        output = dump_bodystructure(fp, r[list(r.keys())[0]]["BODYSTRUCTURE"])
        fp.close()
        self.assertEqual(output, expected)
        return r

    def test_parse_bodystructure(self):
        """Test the parsing of several responses containing BS."""
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_1, """multipart/alternative
    text/plain
    text/html
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_2, """multipart/mixed
    text/plain
    message/rfc822
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_3, """multipart/mixed
    multipart/alternative
        text/plain
        text/html
    application/pdf
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_4, """multipart/mixed
    multipart/alternative
        text/plain
        text/html
    application/octet-stream
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_5, """multipart/alternative
    text/plain
    text/html
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_6, """multipart/mixed
    multipart/related
        multipart/alternative
            text/plain
            text/html
        image/png
        image/jpeg
    application/pdf
    multipart/alternative
        text/plain
        text/html
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_7, """multipart/mixed
    multipart/mixed
        text/plain
    application/octet-stream
""")
        self._test_bodystructure_output(
            data.BODYSTRUCTURE_SAMPLE_8, "text/html\n")



"""Tools to deal with message signatures."""


class EmailSignature(object):
    """User signature.

    :param user: User object
    """

    def __init__(self, user):
        self._sig = u""
        dformat = user.parameters.get_value("editor")
        content = user.parameters.get_value("signature")
        if len(content):
            getattr(self, "_format_sig_%s" % dformat)(content)

    def _format_sig_plain(self, content):
        self._sig = u"""
---
%s""" % content

    def _format_sig_html(self, content):
        content = "---<br>{}".format(content)
        self._sig = content
        return

    def __str__(self):
        return self._sig



# coding: utf-8
"""Declare and register the webmail extension."""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy

from kalabash.core.extensions import KbashExtension, exts_pool
from kalabash.parameters import tools as param_tools

from . import __version__
from . import forms


class Webmail(KbashExtension):
    name = "kalabash_webmail"
    label = "Webmail"
    version = __version__
    description = gettext_lazy("Simple IMAP webmail")
    needs_media = True
    url = "webmail"
    topredirection_url = reverse_lazy("kalabash_webmail:index")

    def load(self):
        param_tools.registry.add("global", forms.ParametersForm, "Webmail")
        param_tools.registry.add("user", forms.UserSettings, "Webmail")


exts_pool.register_extension(Webmail)



import smtplib

from django.core import mail
from django.template.loader import render_to_string

from kalabash.lib.cryptutils import get_password
from kalabash.parameters import tools as param_tools

from . import get_imapconnector, clean_attachments


def send_mail(request, form, posturl=None):
    """Email verification and sending.

    If the form does not present any error, a new MIME message is
    constructed. Then, a connection is established with the defined
    SMTP server and the message is finally sent.

    :param request: a Request object
    :param posturl: the url to post the message form to
    :return: a 2-uple (True|False, HttpResponse)
    """
    if not form.is_valid():
        editormode = request.user.parameters.get_value("editor")
        listing = render_to_string(
            "kalabash_webmail/compose.html",
            {"form": form, "noerrors": True,
             "body": form.cleaned_data.get("body", "").strip(),
             "posturl": posturl},
            request
        )
        return False, {"status": "ko", "listing": listing, "editor": editormode}

    msg = form.to_msg(request)
    conf = dict(param_tools.get_global_parameters("kalabash_webmail"))
    options = {
        "host": conf["smtp_server"], "port": conf["smtp_port"]
    }
    if conf["smtp_secured_mode"] == "ssl":
        options.update({"use_ssl": True})
    elif conf["smtp_secured_mode"] == "starttls":
        options.update({"use_tls": True})
    if conf["smtp_authentication"]:
        options.update({
            "username": request.user.username,
            "password": get_password(request)
        })
    try:
        with mail.get_connection(**options) as connection:
            msg.connection = connection
            msg.send()
    except smtplib.SMTPResponseException as inst:
        return False, {"status": "ko", "error": inst.smtp_error}
    except smtplib.SMTPRecipientsRefused as inst:
        error = ", ".join(["{}: {}".format(rcpt, error)
                           for rcpt, error in inst.recipients.items()])
        return False, {"status": "ko", "error": error}

    # Copy message to sent folder
    sentfolder = request.user.parameters.get_value("sent_folder")
    get_imapconnector(request).push_mail(sentfolder, msg.message())
    clean_attachments(request.session["compose_mail"]["attachments"])
    del request.session["compose_mail"]

    return True, {}



"""AppConfig for webmail."""

from django.apps import AppConfig


class WebmailConfig(AppConfig):
    """App configuration."""

    name = "kalabash_webmail"
    verbose_name = "Simple webmail for Kalabash"

    def ready(self):
        from . import handlers



"""
Set of classes to manipulate/display emails inside the webmail.
"""
import os
import re
import email

import chardet
import six

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.urls import reverse
from django.utils.encoding import smart_str
from django.utils.html import conditional_escape
from django.utils.translation import gettext as _

from kalabash.core.extensions import exts_pool
from kalabash.lib import u2u_decode
from kalabash.lib.email_utils import Email, EmailAddress

from . import imapheader
from .attachments import get_storage_path
from .imaputils import (
    get_imapconnector, BodyStructure
)
from .utils import decode_payload


class ImapEmail(Email):

    """
    A class to represent an email fetched from an IMAP server.
    """

    headernames = [
        ('From', True),
        ('To', True),
        ('Cc', True),
        ('Date', True),
        ('Subject', True),
    ]

    def __init__(self, request, *args, **kwargs):
        super(ImapEmail, self).__init__(*args, **kwargs)
        self.request = request
        self.imapc = get_imapconnector(request)
        self.mbox, self.mailid = self.mailid.split(":")

    def _insert_contact_links(self, addresses):
        """Insert 'add to address book' links."""
        result = []
        title = _("Add to contacts")
        url = reverse("api:contact-list")
        link_tpl = (
            u" <a class='addcontact' href='{}' title='{}'>"
            u"<span class='fa fa-vcard'></span></a>"
        )
        for address in addresses:
            address += link_tpl.format(url, title)
            result.append(address)
        return result

    def fetch_headers(self, raw_addresses=False):
        """Fetch message headers from server."""
        msg = self.imapc.fetchmail(
            self.mbox, self.mailid, readonly=False,
            what=" ".join(self.headers_as_list)
        )
        headers = msg["BODY[HEADER.FIELDS ({})]".format(self.headers_as_text)]
        self.fetch_body_structure(msg)
        msg = email.message_from_string(headers)
        contacts_plugin_installed = exts_pool.get_extension("kalabash_contacts")
        headers_with_address = ("From", "To", "Cc", "Reply-To")
        for hdr in self.headernames:
            label = hdr[0]
            hdrvalue = self.get_header(msg, label, raw=raw_addresses)
            if not hdrvalue:
                continue
            safe = False
            if hdr[1]:
                if label in headers_with_address:
                    if contacts_plugin_installed and not raw_addresses:
                        hdrvalue = self._insert_contact_links(hdrvalue)
                    hdrvalue = ", ".join(hdrvalue)
                    safe = True
                self.headers += [
                    {"name": label, "value": hdrvalue, "safe": safe}]
            label = re.sub("-", "_", label)
            setattr(self, label, hdrvalue)

    def get_header(self, msg, hdrname, **kwargs):
        """Look for a particular header.

        We also try to decode the default value.
        """
        hdrvalue = super().get_header(msg, hdrname)
        if hdrname in ["From", "Reply-To"]:
            # Store a raw copy for further use
            setattr(self, "original_{}".format(hdrname.replace("-", "")),
                    hdrvalue)
        if not hdrvalue:
            return ""
        try:
            key = re.sub("-", "_", hdrname).lower()
            hdrvalue = getattr(imapheader, "parse_%s" % key)(
                hdrvalue, **kwargs)
        except AttributeError:
            pass
        return hdrvalue

    def fetch_body_structure(self, msg=None):
        """Fetch BODYSTRUCTURE for email."""
        if msg is None:
            msg = self.imapc.fetchmail(
                self.mbox, self.mailid, readonly=False
            )
        self.bs = BodyStructure(msg["BODYSTRUCTURE"])
        self._find_attachments()
        if self.dformat not in ["plain", "html"]:
            self.dformat = self.request.user.parameters.get_value(
                self.dformat)
        fallback_fmt = "html" if self.dformat == "plain" else "plain"
        self.mformat = (
            self.dformat if self.dformat in self.bs.contents else fallback_fmt)

    @property
    def headers_as_list(self):
        return [hdr[0].upper() for hdr in self.headernames]

    @property
    def headers_as_text(self):
        return " ".join(self.headers_as_list)

    @property
    def body(self):
        """Load email's body.

        This operation has to be made "on demand" because it requires
        a communication with the IMAP server.
        """
        if self._body is None:
            self.fetch_body_structure()
            bodyc = u""
            parts = self.bs.contents.get(self.mformat, [])
            for part in parts:
                pnum = part["pnum"]
                data = self.imapc._cmd(
                    "FETCH", self.mailid, "(BODY.PEEK[%s])" % pnum
                )
                if not data or not int(self.mailid) in data:
                    continue
                content = decode_payload(
                    part["encoding"], data[int(self.mailid)]["BODY[%s]" % pnum]
                )
                if not isinstance(content, six.text_type):
                    charset = self._find_content_charset(part)
                    if charset is not None:
                        try:
                            content = content.decode(charset)
                        except (UnicodeDecodeError, LookupError):
                            result = chardet.detect(content)
                            content = content.decode(result["encoding"])
                bodyc += content
            self._fetch_inlines()
            if len(bodyc) != 0:
                bodyc = getattr(self, "_post_process_%s" % self.mformat)(bodyc)
                self._body = getattr(self, "viewmail_%s" % self.mformat)(
                    bodyc, links=self.links
                )
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def source(self):
        """Retrieve email source."""
        return self.imapc.fetchmail(
            self.mbox, self.mailid, what="source")['BODY[]']

    def _find_content_charset(self, part):
        for pos, elem in enumerate(part["params"]):
            if elem == "charset":
                return part["params"][pos + 1]
        return None

    def _find_attachments(self):
        """Retrieve attachments from the parsed body structure.

        We try to find and decode a file name for each attachment. If
        we failed, a generic name will be used (ie. part_1, part_2, ...).
        """
        for att in self.bs.attachments:
            attname = "part_{}".format(att["pnum"])
            if "params" in att and att["params"] != "NIL":
                for pos, value in enumerate(att["params"]):
                    if not value.startswith("name"):
                        continue
                    attname = (
                        u2u_decode.u2u_decode(att["params"][pos + 1])
                        .strip("\r\t\n")
                    )
                    break
            elif "disposition" in att and len(att["disposition"]) > 1:
                for pos, value in enumerate(att["disposition"][1]):
                    if not value.startswith("filename"):
                        continue
                    attname = u2u_decode.u2u_decode(
                        att["disposition"][1][pos + 1]
                    ).strip("\r\t\n")
                    break
            self.attachments[att["pnum"]] = smart_str(attname)

    def _fetch_inlines(self):
        """Store inline images on filesystem to display them."""
        for cid, params in list(self.bs.inlines.items()):
            if re.search(r"\.\.", cid):
                continue
            fname = "{}_{}".format(self.mailid, cid)
            path = get_storage_path(fname)
            params["fname"] = os.path.join(
                settings.MEDIA_URL,
                os.path.basename(get_storage_path("")),
                fname
            )
            if default_storage.exists(path):
                continue
            pdef, content = self.imapc.fetchpart(
                self.mailid, self.mbox, params["pnum"]
            )
            default_storage.save(
                path, ContentFile(decode_payload(params["encoding"], content)))

    def _map_cid(self, url):
        m = re.match(".*cid:(.+)", url)
        if m:
            if m.group(1) in self.bs.inlines:
                return self.bs.inlines[m.group(1)]["fname"]
        return url

    def fetch_attachment(self, pnum):
        """Fetch an attachment from the IMAP server."""
        return self.imapc.fetchpart(self.mailid, self.mbox, pnum)


class Modifier(ImapEmail):
    """Message modifier."""

    def __init__(self, form, request, *args, **kwargs):
        kwargs["dformat"] = request.user.parameters.get_value("editor")
        super(Modifier, self).__init__(request, *args, **kwargs)
        self.form = form
        self.fetch_headers(raw_addresses=True)
        getattr(self, "_modify_%s" % self.dformat)()

    def _modify_plain(self):
        self.body = re.sub("</?pre>", "", self.body)

    def _modify_html(self):
        if self.dformat == "html" and self.mformat != self.dformat:
            self.body = re.sub("</?pre>", "", self.body)
            self.body = re.sub("\n", "<br>", self.body)

    @property
    def subject(self):
        """Just a shortcut to return a subject in any case."""
        return self.Subject if hasattr(self, "Subject") else ""


class ReplyModifier(Modifier):
    """Modify a message to reply to it."""

    headernames = ImapEmail.headernames + [
        ("Reply-To", True),
        ("Message-ID", False)
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.textheader = u"%s %s" % (self.From, _("wrote:"))
        if self.dformat == "html":
            self.textheader = u"<p>{}</p>".format(self.textheader)
        if hasattr(self, "Message_ID"):
            self.form.fields["origmsgid"].initial = self.Message_ID
        if not hasattr(self, "Reply_To"):
            self.form.fields["to"].initial = self.original_From
        else:
            self.form.fields["to"].initial = self.original_ReplyTo
        if self.request.GET.get("all", "0") == "1":  # reply-all
            self.form.fields["cc"].initial = ""
            toparse = self.To.split(",")
            if hasattr(self, 'Cc'):
                toparse += self.Cc.split(",")
            for addr in toparse:
                tmp = EmailAddress(addr)
                if tmp.address and tmp.address == self.request.user.username:
                    continue
                if self.form.fields["cc"].initial != "":
                    self.form.fields["cc"].initial += ", "
                self.form.fields["cc"].initial += tmp.fulladdress
        m = re.match(r"re\s*:\s*.+", self.subject.lower())
        if m:
            self.form.fields["subject"].initial = self.subject
        else:
            self.form.fields["subject"].initial = "Re: %s" % self.subject

    def _modify_plain(self):
        super(ReplyModifier, self)._modify_plain()
        lines = self.body.split('\n')
        body = ""
        for l in lines:
            if body != "":
                body += "\n"
            body += ">%s" % l
        self.body = body


class ForwardModifier(Modifier):

    """Modify a message so it can be forwarded."""

    def __init__(self, *args, **kwargs):
        super(ForwardModifier, self).__init__(*args, **kwargs)
        self._header()
        self.form.fields["subject"].initial = u"Fwd: %s" % self.subject

    def __getfunc(self, name):
        return getattr(self, "%s_%s" % (name, self.dformat))

    def _header(self):
        self.textheader = u"{}\n".format(self.__getfunc("_header_begin")())
        self.textheader += (
            self.__getfunc("_header_line")(_("Subject"), self.subject))
        self.textheader += (
            self.__getfunc("_header_line")(_("Date"), self.Date))
        for hdr in ["From", "To", "Reply-To"]:
            try:
                key = re.sub("-", "_", hdr)
                value = getattr(self, key)
                self.textheader += (
                    self.__getfunc("_header_line")(_(hdr), value))
            except AttributeError:
                pass
        self.textheader += self.__getfunc("_header_end")()

    def _header_begin_plain(self):
        return u"----- %s -----" % _("Original message")

    def _header_begin_html(self):
        return u"----- %s -----" % _("Original message")

    def _header_line_plain(self, key, value):
        return u"%s: %s\n" % (key, value)

    def _header_line_html(self, key, value):
        return u"<p>%s: %s</p>" % (key, conditional_escape(value))

    def _header_end_plain(self):
        return u"\n"

    def _header_end_html(self):
        return u""



"""Webmail handlers."""

from django.urls import reverse
from django.dispatch import receiver
from django.utils.translation import gettext as _

from kalabash.core import signals as core_signals

from . import exceptions
from . import lib


@receiver(core_signals.extra_user_menu_entries)
def menu(sender, location, user, **kwargs):
    """Return extra menu entry."""
    if location != "top_menu" or not hasattr(user, "mailbox"):
        return []
    return [
        {"name": "webmail",
         "label": _("Webmail"),
         "url": reverse("kalabash_webmail:index")},
    ]


@receiver(core_signals.user_logout)
def userlogout(sender, request, **kwargs):
    """Close IMAP connection."""
    if not hasattr(request.user, "mailbox"):
        return
    try:
        m = lib.IMAPconnector(user=request.user.username,
                              password=request.session["password"])
    except Exception:
        # TODO silent exception are bad : we should at least log it
        return

    # The following statement may fail under Python 2.6...
    try:
        m.logout()
    except exceptions.ImapError:
        pass


@receiver(core_signals.extra_static_content)
def extra_js(sender, caller, st_type, user, **kwargs):
    """Add javascript."""
    if caller != "user_index" or st_type != "js":
        return ""
    return """function toggleSignatureEditor() {
    var editorId = 'id_kalabash_webmail-signature';
    if ($(this).val() === 'html') {
        CKEDITOR.replace(editorId, $('#' + editorId).data('config'));
    } else {
        var instance = CKEDITOR.instances[editorId];
        instance.destroy();
    }
}

$(document).on(
    'change', 'input[name=kalabash_webmail-editor]', toggleSignatureEditor);
$(document).on('preferencesLoaded', function() {
    $('input[name=kalabash_webmail-editor]:checked').change();
});
"""



# coding: utf-8
"""Custom template tags."""

from six.moves.urllib.parse import urlencode

from django import template
from django.urls import reverse
from django.template.loader import render_to_string
from django.utils.encoding import smart_str
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from ..lib import imapheader, separate_mailbox
from .. import constants

register = template.Library()


@register.simple_tag
def viewmail_menu(selection, folder, user, mail_id=None):
    """Menu of the viewmail location."""
    entries = [{
        "name": "back",
        "url": "javascript:history.go(-1)",
        "img": "fa fa-arrow-left",
        "class": "btn-default",
        "label": _("Back")
    }, {
        "name": "reply",
        "url": "action=reply&mbox=%s&mailid=%s" % (folder, mail_id),
        "img": "fa fa-mail-reply",
        "class": "btn-primary",
        "label": _("Reply"),
        "menu": [{
            "name": "replyall",
            "url": "action=reply&mbox=%s&mailid=%s&all=1" % (folder, mail_id),
            "img": "fa fa-mail-reply-all",
            "label": _("Reply all")
        }, {
            "name": "forward",
            "url": "action=forward&mbox=%s&mailid=%s" % (folder, mail_id),
            "img": "fa fa-mail-forward",
            "label": _("Forward")
        }]
    }, {
        "name": "delete",
        "img": "fa fa-trash",
        "class": "btn-danger",
        "url": u"{0}?mbox={1}&selection[]={2}".format(
            reverse("kalabash_webmail:mail_delete"), folder, mail_id),
        "title": _("Delete")
    }, {
        "name": "mark_as_junk",
        "img": "fa fa-fire",
        "class": "btn-warning",
        "url": u"{0}?mbox={1}&selection[]={2}".format(
            reverse("kalabash_webmail:mail_mark_as_junk"), folder, mail_id),
        "title": _("Mark as spam")
    }, {
        "name": "display_options",
        "title": _("Display options"),
        "img": "fa fa-cog",
        "menu": [{
            "name": "activate_links",
            "label": _("Activate links")
        }, {
            "name": "disable_links",
            "label": _("Disable links")
        }, {
            "name": "show_source",
            "label": _("Show source"),
            "url": u"{}?mbox={}&mailid={}".format(
                reverse("kalabash_webmail:mailsource_get"), folder, mail_id)
        }]
    }]
    if folder == user.parameters.get_value("junk_folder"):
        entries[3] = {
            "name": "mark_as_not_junk",
            "img": "fa fa-thumbs-up",
            "class": "btn-success",
            "url": u"{0}?mbox={1}&selection[]={2}".format(
                reverse("kalabash_webmail:mail_mark_as_not_junk"),
                folder, mail_id),
            "title": _("Mark as not spam")
        }
    menu = render_to_string('common/buttons_list.html',
                            {"selection": selection, "entries": entries,
                             "user": user, "extraclasses": "pull-left"})
    return menu


@register.simple_tag
def compose_menu(selection, backurl, user, **kwargs):
    """The menu of the compose action."""
    entries = [
        {"name": "back",
         "url": "javascript:history.go(-2);",
         "img": "fa fa-arrow-left",
         "class": "btn-default",
         "label": _("Back")},
        {"name": "sendmail",
         "url": "",
         "img": "fa fa-send",
         "class": "btn-default btn-primary",
         "label": _("Send")},
    ]
    context = {
        "selection": selection, "entries": entries, "user": user
    }
    context.update(kwargs)
    return render_to_string('kalabash_webmail/compose_menubar.html', context)


@register.simple_tag
def listmailbox_menu(selection, folder, user, **kwargs):
    """The menu of the listmailbox action."""
    entries = [{
        "name": "totrash",
        "title": _("Delete"),
        "class": "btn-danger",
        "img": "fa fa-trash",
        "url": reverse("kalabash_webmail:mail_delete")
    }, {
        "name": "mark_as_junk_multi",
        "img": "fa fa-fire",
        "class": "btn-warning",
        "url": reverse("kalabash_webmail:mail_mark_as_junk"),
        "title": _("Mark as spam")
    }, {
        "name": "actions",
        "label": _("Actions"),
        "class": "btn btn-default",
        "menu": [{
            "name": "mark-read",
            "label": _("Mark as read"),
            "url": u"{0}?status=read".format(
                reverse("kalabash_webmail:mail_mark", args=[folder]))
        }, {
            "name": "mark-unread",
            "label": _("Mark as unread"),
            "url": u"{0}?status=unread".format(
                reverse("kalabash_webmail:mail_mark", args=[folder]))
        }, {
            "name": "mark-flagged",
            "label": _("Mark as flagged"),
            "url": u"{0}?status=flagged".format(
                reverse("kalabash_webmail:mail_mark", args=[folder]))
        }, {
            "name": "mark-unflagged",
            "label": _("Mark as unflagged"),
            "url": u"{0}?status=unflagged".format(
                reverse("kalabash_webmail:mail_mark", args=[folder]))
        }]
    }]
    sort_actions = [{
        "header": True,
        "label": _("Sort by")
    }]
    current_order = kwargs.get("sort_order")
    for order in constants.SORT_ORDERS:
        entry = {
            "name": "sort_by_{}".format(order[0]),
            "label": order[1],
            "url": order[0],
            "class": "sort-order"
        }
        if current_order[1:] == order[0]:
            css = "fa fa-arrow-{}".format(
                "down" if current_order[0] == "-" else "up")
            entry.update({"img": css})
        sort_actions.append(entry)
    entries[2]["menu"] += sort_actions
    if folder == user.parameters.get_value("trash_folder"):
        entries[0]["class"] += " disabled"
        entries[2]["menu"].insert(4, {
            "name": "empty",
            "label": _("Empty folder"),
            "url": u"{0}?name={1}".format(
                reverse("kalabash_webmail:trash_empty"), folder)
        })
    elif folder == user.parameters.get_value("junk_folder"):
        entries[1] = {
            "name": "mark_as_not_junk_multi",
            "img": "fa fa-thumbs-up",
            "class": "btn-success",
            "url": reverse("kalabash_webmail:mail_mark_as_not_junk"),
            "title": _("Mark as not spam")
        }
    return render_to_string('kalabash_webmail/main_action_bar.html', {
        'selection': selection, 'entries': entries, 'user': user, 'css': "nav",
    })


@register.simple_tag
def print_mailboxes(
        tree, selected=None, withunseen=False, selectonly=False,
        hdelimiter='.'):
    """Display a tree of mailboxes and sub-mailboxes.

    :param tree: the mailboxes to display
    """
    result = ""

    for mbox in tree:
        cssclass = ""
        name = mbox["path"] if "sub" in mbox else mbox["name"]
        label = (
            mbox["label"] if "label" in mbox else
            separate_mailbox(mbox["name"], hdelimiter)[0])
        if mbox.get("removed", False):
            cssclass = "disabled"
        elif selected == name:
            cssclass = "active"
        result += "<li name='%s' class='droppable %s'>\n" % (name, cssclass)
        cssclass = ""
        extra_attrs = ""
        if withunseen and "unseen" in mbox:
            label += " (%d)" % mbox["unseen"]
            cssclass += " unseen"
            extra_attrs = ' data-toggle="%d"' % mbox["unseen"]

        if "sub" in mbox:
            if selected is not None and selected != name and selected.count(
                    name):
                ul_state = "visible"
                div_state = "expanded"
            else:
                ul_state = "hidden"
                div_state = "collapsed"
            result += "<div class='clickbox %s'></div>" % div_state

        result += "<a href='%s' class='%s' name='%s'%s>" % (
            "path" in mbox and mbox["path"] or mbox["name"], cssclass,
            'selectfolder' if selectonly else 'loadfolder', extra_attrs
        )

        iclass = mbox["class"] if "class" in mbox \
            else "fa fa-folder"
        result += "<span class='%s'></span> %s</a>" % (iclass, escape(label))

        if "sub" in mbox and mbox["sub"]:
            result += "<ul name='%s' class='nav nav-pills nav-stacked %s'>" % (
                mbox["path"], ul_state) + print_mailboxes(
                    mbox["sub"], selected, withunseen, selectonly, hdelimiter
            ) + "</ul>\n"
        result += "</li>\n"
    return mark_safe(result)


@register.simple_tag
def mboxes_menu():
    """Mailboxes menu."""
    entries = [
        {"name": "newmbox",
         "url": reverse("kalabash_webmail:folder_add"),
         "img": "fa fa-plus",
         "label": _("Create a new folder"),
         "modal": True,
         "modalcb": "webmail.mboxform_cb",
         "closecb": "webmail.mboxform_close",
         "class": "btn-default btn-xs"},
        {"name": "editmbox",
         "url": reverse("kalabash_webmail:folder_change"),
         "img": "fa fa-edit",
         "label": _("Edit the selected folder"),
         "class": "btn-default btn-xs"},
        {"name": "removembox",
         "url": reverse("kalabash_webmail:folder_delete"),
         "img": "fa fa-trash",
         "label": _("Remove the selected folder"),
         "class": "btn-default btn-xs"},
        {"name": "compress",
         "img": "fa fa-compress",
         "label": _("Compress folder"),
         "class": "btn-default btn-xs",
         "url": reverse("kalabash_webmail:folder_compress")}
    ]

    context = {
        "entries": entries,
        "css": "dropdown-menu",
    }
    return render_to_string('common/menu.html', context)


@register.filter
def parse_imap_header(value, header):
    """Simple template tag to display a IMAP header."""
    safe = True
    try:
        value = getattr(imapheader, "parse_%s" % header)(value)
    except AttributeError:
        pass
    if header == "from":
        value = value[0]
    elif header == "subject":
        safe = False
    return value if not safe else mark_safe(value)


@register.simple_tag
def attachment_url(mbox, mail_id, fname, key):
    """Return full download url of an attachment."""
    url = reverse("kalabash_webmail:attachment_get")
    params = {
        "mbox": mbox,
        "mailid": mail_id,
        "fname": smart_str(fname),
        "partnumber": key
    }
    url = "{}?{}".format(url, urlencode(params))
    return url



# coding: utf-8

"""Webmail extension views."""

import base64
import os

from django.conf import settings
from django.urls import reverse
from django.http import HttpResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils.translation import gettext as _, ngettext
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.gzip import gzip_page

from django.contrib.auth.decorators import login_required

from kalabash.admin.lib import needs_mailbox
from kalabash.core.extensions import exts_pool
from kalabash.lib.exceptions import KalabashException, BadRequest
from kalabash.lib.paginator import Paginator
from kalabash.lib.web_utils import (
    ajax_response, render_to_json_response
)
from kalabash.parameters import tools as param_tools

from .exceptions import UnknownAction
from .forms import (
    FolderForm, AttachmentForm, ComposeMailForm, ForwardMailForm
)
from .lib import (
    decode_payload, AttachmentUploadHandler,
    save_attachment, EmailSignature,
    clean_attachments, set_compose_session, send_mail,
    ImapEmail, WebmailNavigationParameters, ReplyModifier, ForwardModifier,
    get_imapconnector, IMAPconnector, separate_mailbox, rfc6266
)
from .templatetags import webmail_tags


def is_ajax(request):
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'


@login_required
@needs_mailbox()
@gzip_page
def getattachment(request):
    """Fetch a message attachment

    FIXME: par manque de caching, le bodystructure du message est
    redemandé pour accéder aux headers de cette pièce jointe.

    :param request: a ``Request`` object
    """
    mbox = request.GET.get("mbox", None)
    mailid = request.GET.get("mailid", None)
    pnum = request.GET.get("partnumber", None)
    fname = request.GET.get("fname", None)
    if not mbox or not mailid or not pnum or not fname:
        raise BadRequest(_("Invalid request"))

    imapc = get_imapconnector(request)
    partdef, payload = imapc.fetchpart(mailid, mbox, pnum)
    resp = HttpResponse(decode_payload(partdef["encoding"], payload))
    resp["Content-Type"] = partdef["Content-Type"]
    resp["Content-Transfer-Encoding"] = partdef["encoding"]
    resp["Content-Disposition"] = rfc6266.build_header(fname)
    if int(partdef["size"]) < 200:
        resp["Content-Length"] = partdef["size"]
    return resp


@login_required
@needs_mailbox()
def move(request):
    for arg in ["msgset", "to"]:
        if arg not in request.GET:
            raise BadRequest(_("Invalid request"))
    mbc = get_imapconnector(request)
    navparams = WebmailNavigationParameters(request)
    mbc.move(request.GET["msgset"], navparams.get('mbox'), request.GET["to"])
    resp = listmailbox(request, navparams.get('mbox'), update_session=False)
    return render_to_json_response(resp)


@login_required
@needs_mailbox()
def delete(request):
    mbox = request.GET.get("mbox", None)
    selection = request.GET.getlist("selection[]", None)
    if mbox is None or selection is None:
        raise BadRequest(_("Invalid request"))
    selection = [item for item in selection if item.isdigit()]
    mbc = get_imapconnector(request)
    mbc.move(",".join(selection), mbox,
             request.user.parameters.get_value("trash_folder"))
    count = len(selection)
    message = ngettext("%(count)d message deleted",
                        "%(count)d messages deleted",
                        count) % {"count": count}
    return render_to_json_response(message)


@login_required
@needs_mailbox()
def mark(request, name):
    status = request.GET.get("status", None)
    ids = request.GET.get("ids", None)
    if status is None or ids is None:
        raise BadRequest(_("Invalid request"))
    imapc = get_imapconnector(request)
    try:
        getattr(imapc, "mark_messages_%s" % status)(name, ids)
    except AttributeError:
        raise UnknownAction

    return render_to_json_response({
        'action': status, 'mbox': name,
        'unseen': imapc.unseen_messages(name)
    })


def _move_selection_to_folder(request, folder):
    """Move selected messages to the given folder."""
    mbox = request.GET.get("mbox")
    selection = request.GET.getlist("selection[]")
    if mbox is None or selection is None:
        raise BadRequest(_("Invalid request"))
    selection = [item for item in selection if item.isdigit()]
    mbc = get_imapconnector(request)
    mbc.move(",".join(selection), mbox, folder)
    return len(selection)


@login_required
@needs_mailbox()
def mark_as_junk(request):
    """Mark a message as SPAM."""
    count = _move_selection_to_folder(
        request, request.user.parameters.get_value("junk_folder"))
    message = ngettext("%(count)d message marked",
                        "%(count)d messages marked",
                        count) % {"count": count}
    return render_to_json_response(message)


@login_required
@needs_mailbox()
def mark_as_not_junk(request):
    """Mark a message as not SPAM."""
    count = _move_selection_to_folder(request, "INBOX")
    message = ngettext("%(count)d message marked",
                        "%(count)d messages marked",
                        count) % {"count": count}
    return render_to_json_response(message)


@login_required
@needs_mailbox()
def empty(request):
    """Empty the trash folder."""
    name = request.GET.get("name", None)
    if name != request.user.parameters.get_value("trash_folder"):
        raise BadRequest(_("Invalid request"))
    get_imapconnector(request).empty(name)
    content = u"<div class='alert alert-info'>%s</div>" % _("Empty folder")
    return render_to_json_response({
        'listing': content, 'mailbox': name, 'pages': [1]
    })


@login_required
@needs_mailbox()
def folder_compress(request):
    """Compress a mailbox."""
    name = request.GET.get("name", None)
    if name is None:
        raise BadRequest(_("Invalid request"))
    imapc = get_imapconnector(request)
    imapc.compact(name)
    return render_to_json_response({})


@login_required
@needs_mailbox()
def newfolder(request, tplname="kalabash_webmail/folder.html"):
    mbc = IMAPconnector(user=request.user.username,
                        password=request.session["password"])

    if request.method == "POST":
        form = FolderForm(request.POST)
        if form.is_valid():
            pf = request.POST.get("parent_folder", None)
            mbc.create_folder(form.cleaned_data["name"], pf)
            return render_to_json_response({
                'respmsg': _("Folder created"),
                'newmb': form.cleaned_data["name"], 'parent': pf
            })

        return render_to_json_response(
            {'form_errors': form.errors}, status=400)

    ctx = {"title": _("Create a new folder"),
           "formid": "mboxform",
           "action": reverse("kalabash_webmail:folder_add"),
           "action_label": _("Create"),
           "action_classes": "submit",
           "withunseen": False,
           "selectonly": True,
           "mboxes": mbc.getmboxes(request.user),
           "hdelimiter": mbc.hdelimiter,
           "form": FolderForm(),
           "selected": None}
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def editfolder(request, tplname="kalabash_webmail/folder.html"):
    mbc = IMAPconnector(user=request.user.username,
                        password=request.session["password"])
    ctx = {"title": _("Edit folder"),
           "formid": "mboxform",
           "action": reverse("kalabash_webmail:folder_change"),
           "action_label": _("Update"),
           "action_classes": "submit",
           "withunseen": False,
           "selectonly": True,
           "hdelimiter": mbc.hdelimiter}

    if request.method == "POST":
        form = FolderForm(request.POST)
        if form.is_valid():
            pf = request.POST.get("parent_folder", None)
            oldname, oldparent = separate_mailbox(
                request.POST["oldname"], sep=mbc.hdelimiter
            )
            res = {'respmsg': _("Folder updated")}
            if form.cleaned_data["name"] != oldname \
                    or (pf != oldparent):
                newname = form.cleaned_data["name"] if pf is None \
                    else mbc.hdelimiter.join([pf, form.cleaned_data["name"]])
                mbc.rename_folder(request.POST["oldname"], newname)
                res["oldmb"] = oldname
                res["newmb"] = form.cleaned_data["name"]
                res["oldparent"] = oldparent
                res["newparent"] = pf
                WebmailNavigationParameters(request).remove('mbox')
            return render_to_json_response(res)

        return render_to_json_response(
            {'form_errors': form.errors}, status=400)

    name = request.GET.get("name")
    if name is None:
        raise BadRequest(_("Invalid request"))
    shortname, parent = separate_mailbox(name, sep=mbc.hdelimiter)
    ctx = {"title": _("Edit folder"),
           "formid": "mboxform",
           "action": reverse("kalabash_webmail:folder_change"),
           "action_label": _("Update"),
           "action_classes": "submit",
           "withunseen": False,
           "selectonly": True,
           "hdelimiter": mbc.hdelimiter,
           "mboxes": mbc.getmboxes(request.user, until_mailbox=parent),
           "form": FolderForm(),
           "selected": parent}
    ctx["form"].fields["oldname"].initial = name
    ctx["form"].fields["name"].initial = shortname
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def delfolder(request):
    name = request.GET.get("name", None)
    if name is None:
        raise BadRequest(_("Invalid request"))
    mbc = IMAPconnector(user=request.user.username,
                        password=request.session["password"])
    mbc.delete_folder(name)
    WebmailNavigationParameters(request).remove('mbox')
    return ajax_response(request)


@login_required
@csrf_exempt
@needs_mailbox()
def attachments(request, tplname="kalabash_webmail/attachments.html"):
    if request.method == "POST":
        csuploader = AttachmentUploadHandler()
        request.upload_handlers.insert(0, csuploader)
        error = None
        form = AttachmentForm(request.POST, request.FILES)
        if form.is_valid():
            try:
                fobj = request.FILES["attachment"]
                tmpname = save_attachment(fobj)
                request.session["compose_mail"]["attachments"] \
                    += [{"fname": str(fobj),
                         "content-type": fobj.content_type,
                         "size": fobj.size,
                         "tmpname": os.path.basename(tmpname)}]
                request.session.modified = True
                return render(request, "kalabash_webmail/upload_done.html", {
                    "status": "ok", "fname": request.FILES["attachment"],
                    "tmpname": os.path.basename(tmpname)
                })
            except KalabashException as inst:
                error = _("Failed to save attachment: ") + str(inst)

        if csuploader.toobig:
            error = (
                _("Attachment is too big (limit: %s)") %
                param_tools.get_global_parameter("max_attachment_size"))
        return render(request, "kalabash_webmail/upload_done.html", {
            "status": "ko", "error": error
        })
    ctx = {
        "title": _("Attachments"),
        "formid": "uploadfile",
        "target": "upload_target",
        "enctype": "multipart/form-data",
        "form": AttachmentForm(),
        "action": reverse("kalabash_webmail:attachment_list"),
        "attachments": request.session["compose_mail"]["attachments"]
    }
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def delattachment(request):
    """Delete an attachment."""
    name = request.GET.get("name")
    if not name or "compose_mail" not in request.session:
        return ajax_response(request, "ko", respmsg=_("Bad query"))

    error = None
    for att in request.session["compose_mail"]["attachments"]:
        if att["tmpname"] == name:
            request.session["compose_mail"]["attachments"].remove(att)
            fullpath = os.path.join(
                settings.MEDIA_ROOT, "webmail", att["tmpname"]
            )
            try:
                os.remove(fullpath)
            except OSError as e:
                error = _("Failed to remove attachment: ") + str(e)
                break
            request.session.modified = True
            return ajax_response(request)
    if error is None:
        error = _("Unknown attachment")
    return ajax_response(request, "ko", respmsg=error)


def render_mboxes_list(request, imapc):
    """Return the HTML representation of a mailboxes list

    :param request: a ``Request`` object
    :param imapc: an ``IMAPconnector` object
    :return: a string
    """
    curmbox = WebmailNavigationParameters(request).get("mbox", "INBOX")
    return render_to_string("kalabash_webmail/folders.html", {
        "selected": curmbox,
        "mboxes": imapc.getmboxes(request.user),
        "withunseen": True
    }, request)


def listmailbox(request, defmailbox="INBOX", update_session=True):
    """Mailbox content listing.

    Return a list of messages contained in the specified mailbox. The
    number of elements returned depends on the ``MESSAGES_PER_PAGE``
    parameter. (user preferences)

    :param request: a ``Request`` object
    :param defmailbox: the default mailbox (when not present inside
                       request arguments)
    :return: a dictionnary
    """
    navparams = WebmailNavigationParameters(request, defmailbox)
    previous_page_id = int(navparams["page"]) if "page" in navparams else None
    if update_session:
        navparams.store()
    mbox = navparams.get('mbox')
    page_id = int(navparams["page"])
    mbc = get_imapconnector(request)
    mbc.parse_search_parameters(
        navparams.get("criteria"), navparams.get("pattern"))
    sort_order = navparams.get("order")
    paginator = Paginator(
        mbc.messages_count(folder=mbox, order=sort_order),
        request.user.parameters.get_value("messages_per_page")
    )
    page = paginator.getpage(page_id)
    content = ""
    if page is not None:
        email_list = mbc.fetch(page.id_start, page.id_stop, mbox)
        content = render_to_string(
            "kalabash_webmail/email_list.html", {
                "email_list": email_list,
                "page": page_id,
                "with_top_div": request.GET.get("scroll", "false") == "false"
            }, request
        )
        length = len(content)
    else:
        if page_id == 1:
            content = u"<div class='alert alert-info'>{0}</div>".format(
                _("Empty mailbox")
            )
        length = 0
        if previous_page_id is not None:
            navparams["page"] = previous_page_id
    return {
        "listing": content, "length": length, "pages": [page_id],
        "menuargs": {"sort_order": sort_order}
    }


def render_compose(request, form, posturl, email=None, **kwargs):
    """Render the compose form."""
    resp = {}
    if email is None:
        body = ""
        textheader = ""
    else:
        body = email.body
        textheader = email.textheader
    if kwargs.get("insert_signature"):
        signature = EmailSignature(request.user)
        body = "{}{}".format(body, signature)
    randid = None
    condition = (
        "id" not in request.GET or
        "compose_mail" not in request.session or
        request.session["compose_mail"]["id"] != request.GET["id"]
    )
    if condition:
        randid = set_compose_session(request)
        if kwargs.get("load_email_attachments"):
            for pnum, fname in email.attachments.items():
                partdef, payload = email.fetch_attachment(pnum)
                request.session["compose_mail"]["attachments"].append({
                    "fname": fname,
                    "content-type": partdef["Content-Type"],
                    "size": partdef["size"],
                    "tmpname": save_attachment(base64.b64decode(payload))
                })
            request.session.modified = True

    attachment_list = request.session["compose_mail"]["attachments"]
    if attachment_list:
        resp["menuargs"] = {"attachment_counter": len(attachment_list)}

    if textheader:
        body = "{}\n{}".format(textheader, body)
    form.fields["body"].initial = body
    content = render_to_string("kalabash_webmail/compose.html", {
        "form": form, "posturl": posturl
    }, request)

    resp.update({
        "listing": content,
        "editor": request.user.parameters.get_value("editor")
    })
    if randid is not None:
        resp["id"] = randid
    return resp


def compose(request):
    """Compose email."""
    url = "?action=compose"
    if request.method == "POST":
        form = ComposeMailForm(request.user, request.POST)
        status, resp = send_mail(request, form, posturl=url)
        return resp

    form = ComposeMailForm(request.user)
    return render_compose(request, form, url, insert_signature=True)


def get_mail_info(request):
    """Retrieve a mailbox and an email ID from a request.
    """
    mbox = request.GET.get("mbox", None)
    mailid = request.GET.get("mailid", None)
    if mbox is None or mailid is None:
        raise BadRequest(_("Invalid request"))
    return mbox, mailid


def new_compose_form(request, action, mbox, mailid, **kwargs):
    """Return a new composition form.

    Valid for reply and forward actions only.
    """
    form = ComposeMailForm(request.user)
    modclass = globals()["%sModifier" % action.capitalize()]
    email = modclass(form, request, "%s:%s" % (mbox, mailid), links=True)
    url = "?action=%s&mbox=%s&mailid=%s" % (action, mbox, mailid)
    return render_compose(request, form, url, email, **kwargs)


def reply(request):
    """Reply to email."""
    mbox, mailid = get_mail_info(request)
    if request.method == "POST":
        url = "?action=reply&mbox=%s&mailid=%s" % (mbox, mailid)
        form = ComposeMailForm(request.user, request.POST)
        status, resp = send_mail(request, form, url)
        if status:
            get_imapconnector(request).msg_answered(mbox, mailid)
        return resp
    return new_compose_form(request, "reply", mbox, mailid)


def forward(request):
    """Forward email."""
    mbox, mailid = get_mail_info(request)
    if request.method == "POST":
        url = "?action=forward&mbox=%s&mailid=%s" % (mbox, mailid)
        form = ForwardMailForm(request.user, request.POST)
        status, resp = send_mail(request, form, url)
        if status:
            get_imapconnector(request).msg_forwarded(mbox, mailid)
        return resp
    return new_compose_form(
        request, "forward", mbox, mailid, load_email_attachments=True)


@login_required
@needs_mailbox()
def getmailcontent(request):
    mbox = request.GET.get("mbox", None)
    mailid = request.GET.get("mailid", None)
    if mbox is None or mailid is None:
        raise BadRequest(_("Invalid request"))
    email = ImapEmail(
        request,
        "%s:%s" % (mbox, mailid), dformat="DISPLAYMODE",
        links=request.GET.get("links", "0") == "1"
    )
    return render(request, "common/viewmail.html", {
        "mailbody": email.body if email.body else ""
    })


@login_required
@needs_mailbox()
def getmailsource(request):
    """Retrieve message source."""
    mbox = request.GET.get("mbox", None)
    mailid = request.GET.get("mailid", None)
    if mbox is None or mailid is None:
        raise BadRequest(_("Invalid request"))
    email = ImapEmail(
        request,
        "%s:%s" % (mbox, mailid), dformat="DISPLAYMODE",
        links=request.GET.get("links", "0") == "1"
    )
    return render(request, "kalabash_webmail/mail_source.html", {
        "title": _("Message source"),
        "source": email.source
    })


def viewmail(request):
    mbox = request.GET.get("mbox", None)
    mailid = request.GET.get("mailid", None)
    if mbox is None or mailid is None:
        raise BadRequest(_("Invalid request"))
    links = request.GET.get("links", None)
    if links is None:
        links = int(request.user.parameters.get_value("enable_links"))
    else:
        links = int(links)
    email = ImapEmail(
        request,
        "%s:%s" % (mbox, mailid), dformat="DISPLAYMODE", links=links
    )
    email.fetch_headers()
    context = {
        "mbox": mbox,
        "mailid": mailid,
        "links": links,
        "headers": email.headers,
        "attachments": email.attachments
    }
    content = render_to_string(
        "kalabash_webmail/headers.html", context, request)
    return {"listing": content, "menuargs": {"mail_id": mailid}}


@login_required
@needs_mailbox()
def submailboxes(request):
    """Retrieve the sub mailboxes of a mailbox."""
    topmailbox = request.GET.get('topmailbox', '')
    with_unseen = request.GET.get('unseen', None)
    mboxes = get_imapconnector(request).getmboxes(
        request.user, topmailbox, unseen_messages=with_unseen == 'true')
    return render_to_json_response(mboxes)


@login_required
@needs_mailbox()
def check_unseen_messages(request):
    mboxes = request.GET.get("mboxes", None)
    if not mboxes:
        raise BadRequest(_("Invalid request"))
    mboxes = mboxes.split(",")
    counters = {}
    imapc = get_imapconnector(request)
    for mb in mboxes:
        counters[mb] = imapc.unseen_messages(mb)
    return render_to_json_response(counters)


@login_required
@needs_mailbox()
def index(request):
    """Webmail actions handler

    Problèmes liés à la navigation 'anchor based'
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    Lors d'un rafraichissemt complet, une première requête est envoyée
    vers /kalabash_webmail/. On ne connait pas encore l'action qui va être
    demandée mais on peut déjà envoyer des informations indépendantes
    (comme les dossiers, le quota).

    Si on se contente de cela, l'affichage donnera un aspect décomposé
    qui n'est pas très séduisant (à cause de la latence notamment). Il
    faudrait pouvoir envoyer le menu par la même occasion, le souci
    étant de savoir lequel...

    Une solution possible : il suffirait de déplacer le menu vers la
    droite pour l'aligner avec le contenu, remonter la liste des
    dossiers (même hauteur que le menu) et renvoyer le menu en même
    temps que le contenu. Le rendu sera plus uniforme je pense.

    """
    action = request.GET.get("action", None)
    if action is not None:
        if action not in globals():
            raise UnknownAction
        response = globals()[action](request)
    else:
        if is_ajax(request):
            raise BadRequest(_("Invalid request"))
        response = {"selection": "webmail"}

    curmbox = WebmailNavigationParameters(request).get("mbox", "INBOX")
    if not is_ajax(request):
        request.session["lastaction"] = None
        imapc = get_imapconnector(request)
        imapc.getquota(curmbox)
        trash = request.user.parameters.get_value("trash_folder")
        response.update({
            "hdelimiter": imapc.hdelimiter,
            "mboxes": render_mboxes_list(request, imapc),
            "refreshrate": request.user.parameters.get_value(
                "refresh_interval"),
            "quota": imapc.quota_usage,
            "trash": trash,
            "ro_mboxes": [
                "INBOX", "Junk",
                request.user.parameters.get_value("sent_folder"),
                trash,
                request.user.parameters.get_value("drafts_folder")
            ],
            "mboxes_col_width": request.user.parameters.get_value(
                "mboxes_col_width"),
            "contacts_plugin_enabled": exts_pool.get_extension(
                "kalabash_contacts")
        })
        return render(request, "kalabash_webmail/index.html", response)

    if action in ["reply", "forward"]:
        action = "compose"
    if request.session["lastaction"] != action:
        extra_args = {}
        if "menuargs" in response:
            extra_args = response["menuargs"]
            del response["menuargs"]
        try:
            menu = getattr(webmail_tags, "%s_menu" % action)
            response["menu"] = menu("", curmbox, request.user, **extra_args)
        except KeyError:
            pass

    response.update(callback=action)
    http_status = 200
    if "status" in response:
        del response['status']
        http_status = 400
    return render_to_json_response(response, status=http_status)



"""Webmail constants."""

from django.utils.translation import gettext_lazy as _


SORT_ORDERS = [
    ("date", _("Date")),
    ("from", _("Sender")),
    ("size", _("Size")),
    ("subject", _("Subject")),
]



# coding: utf-8

"""Simple parser for FETCH responses.

The ``imaplib`` module doesn't parse IMAP responses, it returns raw
values. Since Kalabash relies on BODYSTRUCTURE attributes to display
messages (we don't want to overload the server), a parser is required.

Python 2/3 compatibility note: this parser excepts bytes objects when
run with Python3 and str (not unicode) ones with Python2.
"""

from __future__ import print_function

import re

import chardet
import six


class ParseError(Exception):
    """Generic parsing error"""
    pass


class Lexer(object):
    """The lexical analysis part.

    This class provides a simple way to define tokens (with patterns)
    to be detected. Patterns are provided using a list of 2-uple. Each
    2-uple consists of a token name and an associated pattern.

    Example: [("left_bracket", r'\['),]
    """

    def __init__(self, definitions):
        self.definitions = definitions
        parts = []
        for name, part in definitions:
            parts.append(r"(?P<%s>%s)" % (name, part))
        self.regexpString = "|".join(parts)
        self.regexp = re.compile(self.regexpString, re.MULTILINE)
        self.wsregexp = re.compile(r"\s+", re.M)

    def curlineno(self):
        """Return the current line number"""
        return self.text[:self.pos].count("\n") + 1

    def scan(self, text):
        """Analyze some data.

        Analyse the passed content. Each time a token is recognized, a
        2-uple containing its name and the parsed value is raised
        (using yield).

        :param text: a string containing the data to parse
        :raises: ParseError
        """
        self.pos = 0
        self.text = text
        while self.pos < len(text):
            m = self.wsregexp.match(text, self.pos)
            if m is not None:
                self.pos = m.end()
                continue

            m = self.regexp.match(text, self.pos)
            if m is None:
                raise ParseError("unknown token {}".format(text[self.pos:]))

            self.pos = m.end()
            yield (m.lastgroup, m.group(m.lastgroup))


class FetchResponseParser(object):
    """A token generator.

    By *token*, I mean: *literal*, *quoted* or anything else until the
    next ' ' or ')' character (number, NIL and others should fall into
    this last category).
    """

    rules = [
        ("left_parenthesis", r'\('),
        ("right_parenthesis", r'\)'),
        ("string", r'"([^"\\]|\\.)*"'),
        ("nil", r'NIL'),
        ("data_item",
         r"(?P<name>[A-Z][A-Z\.0-9]+)"
         r"(?P<section>\[.*\])?(?P<origin_octet>\<\d+\>)?"),
        ("number", r'[0-9]+'),
        ("literal_marker", r'{\d+}'),
        ("flag", r'(\\|\$)?[a-zA-Z0-9\-_]+'),
    ]

    def __init__(self):
        """Constructor."""
        self.lexer = Lexer(self.rules)
        self.__reset_parser()

    def __reset_parser(self):
        """Reset parser states."""
        self.result = {}
        self.__current_message = {}
        self.__next_literal_len = 0
        self.__cur_data_item = None
        self.__args_parsing_func = None
        self.__expected = None
        self.__depth = 0
        self.__bs_stack = []

    def set_expected(self, *args):
        """Indicate next expected token types."""
        self.__expected = args

    def __default_args_parser(self, ttype, tvalue):
        """Default arguments parser."""
        self.__current_message[self.__cur_data_item] = tvalue
        self.__args_parsing_func = None

    def __flags_args_parser(self, ttype, tvalue):
        """FLAGS arguments parser."""
        if ttype == "left_parenthesis":
            self.__current_message[self.__cur_data_item] = []
            self.__depth += 1
        elif ttype == "flag":
            self.__current_message[self.__cur_data_item].append(tvalue)
            self.set_expected("flag", "right_parenthesis")
        elif ttype == "right_parenthesis":
            self.__args_parsing_func = None
            self.__depth -= 1
        else:
            raise ParseError(
                "Unexpected token found: {}".format(ttype))

    def __set_part_numbers(self, bs, prefix=""):
        """Set part numbers."""
        cpt = 1
        for mp in bs:
            if isinstance(mp, list):
                self.__set_part_numbers(mp, prefix)
            elif isinstance(mp, dict):
                if isinstance(mp["struct"][0], list):
                    nprefix = "{}{}.".format(prefix, cpt)
                    self.__set_part_numbers(mp["struct"][0], nprefix)
                mp["partnum"] = "{}{}".format(prefix, cpt)
                cpt += 1

    def __bstruct_args_parser(self, ttype, tvalue):
        """BODYSTRUCTURE arguments parser."""
        if ttype == "left_parenthesis":
            self.__bs_stack = [[]] + self.__bs_stack
            return
        if ttype == "right_parenthesis":
            if len(self.__bs_stack) > 1:
                part = self.__bs_stack.pop(0)
                # Check if we are parsing a list of mime part or a
                # list or arguments.
                condition = (
                    len(self.__bs_stack[0]) > 0 and
                    not isinstance(self.__bs_stack[0][0], dict))
                if condition:
                    self.__bs_stack[0].append(part)
                else:
                    self.__bs_stack[0].append({"struct": part})
            else:
                # End of BODYSTRUCTURE
                if not isinstance(self.__bs_stack[0][0], list):
                    # Special case for non multipart structures
                    self.__bs_stack[0] = {"struct": self.__bs_stack[0]}
                self.__set_part_numbers(self.__bs_stack)
                self.__current_message[self.__cur_data_item] = self.__bs_stack
                self.__bs_stack = []
                self.__args_parsing_func = None
            return
        if ttype == "string":
            tvalue = tvalue.strip('"')
            # Check if previous element was a mime part. If so, we are
            # dealing with a 'multipart' mime part...
            condition = (
                len(self.__bs_stack[0]) and
                isinstance(self.__bs_stack[0][-1], dict))
            if condition:
                self.__bs_stack[0] = [self.__bs_stack[0]] + [tvalue]
                return
        elif ttype == "number":
            tvalue = int(tvalue)
        self.__bs_stack[0].append(tvalue)

    def __parse_data_item(self, ttype, tvalue):
        """Find next data item."""
        if ttype == "data_item":
            self.__cur_data_item = tvalue
            if tvalue == "BODYSTRUCTURE":
                self.set_expected("left_parenthesis")
                self.__args_parsing_func = self.__bstruct_args_parser
            elif tvalue == "FLAGS":
                self.set_expected("left_parenthesis")
                self.__args_parsing_func = self.__flags_args_parser
            else:
                self.__args_parsing_func = self.__default_args_parser
            return
        elif ttype == "right_parenthesis":
            self.__depth -= 1
            assert self.__depth == 0
            # FIXME: sometimes, FLAGS are returned outside the UID
            # scope (see sample 1 in tests). For now, we just ignore
            # them but we need a better solution!
            if "UID" in self.__current_message:
                self.result[int(self.__current_message.pop("UID"))] = (
                    self.__current_message)
            self.__current_message = {}
            return
        raise ParseError(
            "unexpected {} found while looking for data_item near {}"
            .format(ttype, tvalue))

    def __convert_to_str(self, chunk):
        """Convert chunk to str and guess encoding."""
        condition = (
            six.PY2 and isinstance(chunk, six.text_type) or
            six.PY3 and isinstance(chunk, six.binary_type)
        )
        if not condition:
            return chunk
        try:
            chunk = chunk.decode("utf-8")
        except UnicodeDecodeError:
            pass
        else:
            return chunk
        try:
            result = chardet.detect(chunk)
        except UnicodeDecodeError:
            raise RuntimeError("Can't find string encoding")
        return chunk.decode(result["encoding"])

    def parse_chunk(self, chunk):
        """Parse chunk."""
        if not chunk:
            return
        chunk = self.__convert_to_str(chunk)
        if self.__next_literal_len:
            literal = chunk[:self.__next_literal_len]
            chunk = chunk[self.__next_literal_len:]
            self.__next_literal_len = 0
            if self.__cur_data_item != "BODYSTRUCTURE":
                self.__current_message[self.__cur_data_item] = literal
                self.__args_parsing_func = None
            else:
                self.__args_parsing_func("literal", literal)
        for ttype, tvalue in self.lexer.scan(chunk):
            if self.__expected is not None:
                if ttype not in self.__expected:
                    raise ParseError(
                        "unexpected {} found while looking for {}"
                        .format(ttype, "|".join(self.__expected)))
                self.__expected = None
            if ttype == "literal_marker":
                self.__next_literal_len = int(tvalue[1:-1])
                continue
            elif self.__depth == 0:
                if ttype == "number":
                    # We should start here with a message ID
                    self.set_expected("left_parenthesis")
                if ttype == "left_parenthesis":
                    self.__depth += 1
                continue
            elif self.__args_parsing_func is None:
                self.__parse_data_item(ttype, tvalue)
                continue
            self.__args_parsing_func(ttype, tvalue)

    def parse(self, data):
        """Parse received data."""
        self.__reset_parser()
        for chunk in data:
            if isinstance(chunk, tuple):
                for schunk in chunk:
                    self.parse_chunk(schunk)
            else:
                self.parse_chunk(chunk)
        return self.result



"""
Set of functions used to parse and transform email headers.
"""

from __future__ import unicode_literals

import datetime
import email

import chardet
import six

from django.utils import timezone
from django.utils.html import escape
from django.utils.formats import date_format

from kalabash.lib.email_utils import EmailAddress
from kalabash.lib.signals import get_request


__all__ = [
    'parse_from', 'parse_to', 'parse_message_id', 'parse_date',
    'parse_reply_to', 'parse_cc', 'parse_subject'
]

# date and time formats for email list
# according to https://en.wikipedia.org/wiki/Date_format_by_country
# and https://en.wikipedia.org/wiki/Date_and_time_representation_by_country
DATETIME_FORMATS = {
    "cs": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "de": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "en": {'SHORT': 'l, P', 'LONG': 'N j, Y P'},
    "es": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "fr": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "it": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "ja_JP": {'SHORT': 'l, P', 'LONG': 'N j, Y P'},
    "nl": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "pl_PL": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "pt_PT": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "pt_BR": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "ru": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
    "sv": {'SHORT': 'l, H:i', 'LONG': 'd. N Y H:i'},
}


def to_unicode(value):
    """Try to convert a string to unicode."""
    condition = (
        value is None or isinstance(value, six.text_type)
    )
    if condition:
        return value
    try:
        value = value.decode("utf-8")
    except UnicodeDecodeError:
        pass
    else:
        return value
    try:
        res = chardet.detect(value)
    except UnicodeDecodeError:
        return value
    if res["encoding"] == "ascii":
        return value
    return value.decode(res["encoding"])


def parse_address(value, **kwargs):
    """Parse an email address."""
    addr = EmailAddress(value)
    if kwargs.get("raw"):
        return to_unicode(addr.fulladdress)
    if addr.name:
        return u"<span title={}>{}</span>".format(
            to_unicode(addr.address), escape(to_unicode(addr.name)))
    return u"<span>{}</span>".format(to_unicode(addr.address))


def parse_address_list(values, **kwargs):
    """Parse a list of email addresses."""
    lst = values.split(",")
    result = []
    for addr in lst:
        result.append(parse_address(addr, **kwargs))
    return result


def parse_from(value, **kwargs):
    """Parse a From: header."""
    return [parse_address(value, **kwargs)]


def parse_to(value, **kwargs):
    """Parse a To: header."""
    return parse_address_list(value, **kwargs)


def parse_cc(value, **kwargs):
    """Parse a Cc: header."""
    return parse_address_list(value, **kwargs)


def parse_reply_to(value, **kwargs):
    """Parse a Reply-To: header.
    """
    return parse_address_list(value, **kwargs)


def parse_date(value, **kwargs):
    """Parse a Date: header."""
    tmp = email.utils.parsedate_tz(value)
    if not tmp:
        return value
    ndate = datetime.datetime.fromtimestamp(email.utils.mktime_tz(tmp))
    if ndate.tzinfo is not None:
        tz = timezone.get_current_timezone()
        ndate = datetime.datetime.fromtimestamp(ndate).replace(tzinfo=tz)
    current_language = get_request().user.language
    if datetime.datetime.now() - ndate > datetime.timedelta(7):
        fmt = "LONG"
    else:
        fmt = "SHORT"
    return date_format(
        ndate,
        DATETIME_FORMATS.get(current_language, DATETIME_FORMATS.get("en"))[fmt]
    )


def parse_message_id(value, **kwargs):
    """Parse a Message-ID: header."""
    return value.strip('\n')


def parse_subject(value, **kwargs):
    """Parse a Subject: header."""
    from kalabash.lib import u2u_decode

    try:
        subject = u2u_decode.u2u_decode(value)
    except UnicodeDecodeError:
        subject = value
    return to_unicode(subject)



# coding: utf-8
from django.urls import path

from . import views

app_name = 'kalabash_webmail'

urlpatterns = [
    path('', views.index, name="index"),
    path('submailboxes', views.submailboxes, name="submailboxes_get"),
    path('getmailcontent', views.getmailcontent, name="mailcontent_get"),
    path('getmailsource', views.getmailsource, name="mailsource_get"),
    path('unseenmsgs', views.check_unseen_messages,
         name="unseen_messages_check"),

    path('delete/', views.delete, name="mail_delete"),
    path('move/', views.move, name="mail_move"),
    path('mark/<path:name>/', views.mark, name="mail_mark"),
    path('mark_as_junk/', views.mark_as_junk, name="mail_mark_as_junk"),
    path('mark_as_not_junk/', views.mark_as_not_junk,
         name="mail_mark_as_not_junk"),

    path('newfolder/', views.newfolder, name="folder_add"),
    path('editfolder/', views.editfolder, name="folder_change"),
    path('delfolder/', views.delfolder, name="folder_delete"),
    path('compressfolder/', views.folder_compress, name="folder_compress"),
    path('emptytrash/', views.empty, name="trash_empty"),

    path('attachments/', views.attachments, name="attachment_list"),
    path('delattachment/', views.delattachment, name="attachment_delete"),
    path('getattachment/', views.getattachment, name="attachment_get"),
]



# coding: utf-8

# flake8: noqa

"""Tests data."""

BODYSTRUCTURE_SAMPLE_1 = [
    b'36 (FLAGS (\\Seen))',
    b'36 (UID 36 BODYSTRUCTURE (("text" "plain" ("charset" "UTF-8") NIL NIL "QUOTED-PRINTABLE" 959 29 NIL ("inline" NIL) NIL NIL)("text" "html" ("charset" "UTF-8") NIL NIL "QUOTED-PRINTABLE" 14695 322 NIL ("inline" NIL) NIL NIL) "alternative" ("boundary" "----=_Part_2437867_661044267.1501268072105") NIL NIL NIL))'
]

BODYSTRUCTURE_SAMPLE_2 = [
    (b'19 (UID 19 FLAGS (\\Seen $label4 user_flag-1) BODYSTRUCTURE (("text" "plain" ("charset" "ISO-8859-1" "format" "flowed") NIL NIL "7bit" 2 1 NIL NIL NIL NIL)("message" "rfc822" ("name*" "ISO-8859-1\'\'%5B%49%4E%53%43%52%49%50%54%49%4F%4E%5D%20%52%E9%63%E9%70%74%69%6F%6E%20%64%65%20%76%6F%74%72%65%20%64%6F%73%73%69%65%72%20%64%27%69%6E%73%63%72%69%70%74%69%6F%6E%20%46%72%65%65%20%48%61%75%74%20%44%E9%62%69%74") NIL NIL "8bit" 3632 ("Wed, 13 Dec 2006 20:30:02 +0100" {70}',
    b"[INSCRIPTION] R\xe9c\xe9ption de votre dossier d'inscription Free Haut D\xe9bit"),
    (b' (("Free Haut Debit" NIL "inscription" "freetelecom.fr")) (("Free Haut Debit" NIL "inscription" "freetelecom.fr")) ((NIL NIL "hautdebit" "freetelecom.fr")) ((NIL NIL "nguyen.antoine" "wanadoo.fr")) NIL NIL NIL "<20061213193125.9DA0919AC@dgroup2-2.proxad.net>") ("text" "plain" ("charset" "iso-8859-1") NIL NIL "8bit" 1428 38 NIL ("inline" NIL) NIL NIL) 76 NIL ("inline" ("filename*" "ISO-8859-1\'\'%5B%49%4E%53%43%52%49%50%54%49%4F%4E%5D%20%52%E9%63%E9%70%74%69%6F%6E%20%64%65%20%76%6F%74%72%65%20%64%6F%73%73%69%65%72%20%64%27%69%6E%73%63%72%69%70%74%69%6F%6E%20%46%72%65%65%20%48%61%75%74%20%44%E9%62%69%74")) NIL NIL) "mixed" ("boundary" "------------040706080908000209030901") NIL NIL NIL) BODY[HEADER.FIELDS (DATE FROM TO CC SUBJECT)] {266}',
     b'Date: Tue, 19 Dec 2006 19:50:13 +0100\r\nFrom: Alphamonak Solutions <nguyen.antoine@wanadoo.fr>\r\nTo: Alphamonak Solutions <amonak@alphamonak.com>\r\nSubject: [Fwd: [INSCRIPTION] =?ISO-8859-1?Q?R=E9c=E9ption_de_votre_?=\r\n =?ISO-8859-1?Q?dossier_d=27inscription_Free_Haut_D=E9bit=5D?=\r\n\r\n'),
    b')'
]

BODYSTRUCTURE_SAMPLE_3 = [
    (b'58 (UID 123753 BODYSTRUCTURE ((("text" "plain" ("charset" "iso-8859-1") NIL NIL "quoted-printable" 90 10 NIL NIL NIL NIL)("text" "html" ("charset" "iso-8859-1") NIL NIL "quoted-printable" 1034 33 NIL NIL NIL NIL) "alternative" ("boundary" "_000_HE1PR10MB1642F2B8BA7FF8EAC0FC7AECD86E0HE1PR10MB1642EURP_") NIL NIL NIL)("application" "pdf" ("name" "=?iso-8859-1?Q?CV=5FAude=5FGIRODON=5FNGUYEN=5Fao=FBt2017_g=E9n=E9rique.pd?= =?iso-8859-1?Q?f?=") NIL {95}', '=?iso-8859-1?Q?CV=5FAude=5FGIRODON=5FNGUYEN=5Fao=FBt2017_g=E9n=E9rique.pd?=\n =?iso-8859-1?Q?f?='),
    b' "base64" 94130 NIL ("attachment" ("filename" "=?iso-8859-1?Q?CV=5FAude=5FGIRODON=5FNGUYEN=5Fao=FBt2017_g=E9n=E9rique.pd?= =?iso-8859-1?Q?f?=" "size" "68787" "creation-date" "Wed, 13 Sep 2017 08:50:03 GMT" "modification-date" "Wed, 13 Sep 2017 08:50:03 GMT")) NIL NIL) "mixed" ("boundary" "_004_HE1PR10MB1642F2B8BA7FF8EAC0FC7AECD86E0HE1PR10MB1642EURP_") NIL ("fr-FR") NIL))'
]

BODYSTRUCTURE_4 = b'BODYSTRUCTURE ((("text" "plain" ("charset" "iso-8859-1") NIL NIL "quoted-printable" 886 32 NIL NIL NIL NIL)("text" "html" ("charset" "us-ascii") NIL NIL "quoted-printable" 1208 16 NIL NIL NIL NIL) "alternative" ("boundary" "----=_NextPart_001_0003_01CCC564.B2F64FF0") NIL NIL NIL)("application" "octet-stream" ("name" "Carte Verte_2.pdf") NIL NIL "base64" 285610 NIL ("attachment" ("filename" "Carte Verte_2.pdf")) NIL NIL) "mixed" ("boundary" "----=_NextPart_000_0002_01CCC564.B2F64FF0") NIL NIL NIL)'

BODYSTRUCTURE_SAMPLE_4 = [
    (b'855 (UID 46931 ' + BODYSTRUCTURE_4 + b' BODY[HEADER.FIELDS (FROM TO CC DATE SUBJECT REPLY-TO MESSAGE-ID)] {235}', b'From: <Service.client10@maaf.fr>\r\nTo: <TONIO@NGYN.ORG>\r\nCc: \r\nSubject: Notre contact du 28/12/2011 - 192175092\r\nDate: Wed, 28 Dec 2011 13:29:17 +0100\r\nMessage-ID: <CABY0dkJspTaFn7v-1OG1nc9M0Qxn+VUTpcXzxyGNBnSnZtqMrw@mail.gmail.com>\r\n\r\n'),
    b')'
]

BODYSTRUCTURE_ONLY_4 = [
    (b'855 (UID 46931 ' + BODYSTRUCTURE_4), ')']

BODYSTRUCTURE_ONLY_5 = [
    (b'855 (UID 46932 ' + BODYSTRUCTURE_4), ')']

BODY_PLAIN_4 = [
    (b'855 (UID 46931 BODY[1.1] {25}', b'This is a test message.\r\n'), b')']

BODYSTRUCTURE_SAMPLE_5 = [
    (b'856 (UID 46936 BODYSTRUCTURE (("text" "plain" ("charset" "ISO-8859-1") NIL NIL "quoted-printable" 724 22 NIL NIL NIL NIL)("text" "html" ("charset" "ISO-8859-1") NIL NIL "quoted-printable" 2662 48 NIL NIL NIL NIL) "alternative" ("boundary" "----=_Part_1326887_254624357.1325083973970") NIL NIL NIL) BODY[HEADER.FIELDS (DATE FROM TO CC SUBJECT)] {258}', 'Date: Wed, 28 Dec 2011 15:52:53 +0100 (CET)\r\nFrom: =?ISO-8859-1?Q?Malakoff_M=E9d=E9ric?= <communication@communication.malakoffmederic.com>\r\nTo: Alphamonak Solutions <amonak@alphamonak.com>\r\nSubject: =?ISO-8859-1?Q?Votre_inscription_au_grand_Jeu_Malakoff_M=E9d=E9ric?=\r\n\r\n'),
    b')'
]

BODYSTRUCTURE_SAMPLE_6 = [
    (b'123 (UID 3 BODYSTRUCTURE (((("text" "plain" ("charset" "iso-8859-1") NIL NIL "quoted-printable" 1266 30 NIL NIL NIL NIL)("text" "html" ("charset" "iso-8859-1") NIL NIL "quoted-printable" 8830 227 NIL NIL NIL NIL) "alternative" ("boundary" "_000_152AC7ECD1F8AB43A9AD95DBDDCA3118082C09GKIMA24cmcicfr_") NIL NIL NIL)("image" "png" ("name" "image005.png") "<image005.png@01CC6CAA.4FADC490>" "image005.png" "base64" 7464 NIL ("inline" ("filename" "image005.png" "size" "5453" "creation-date" "Tue, 06 Sep 2011 13:33:49 GMT" "modification-date" "Tue, 06 Sep 2011 13:33:49 GMT")) NIL NIL)("image" "jpeg" ("name" "image006.jpg") "<image006.jpg@01CC6CAA.4FADC490>" "image006.jpg" "base64" 2492 NIL ("inline" ("filename" "image006.jpg" "size" "1819" "creation-date" "Tue, 06 Sep 2011 13:33:49 GMT" "modification-date" "Tue, 06 Sep 2011 13:33:49 GMT")) NIL NIL) "related" ("boundary" "_006_152AC7ECD1F8AB43A9AD95DBDDCA3118082C09GKIMA24cmcicfr_" "type" "multipart/alternative") NIL NIL NIL)("application" "pdf" ("name" "bilan assurance CIC.PDF") NIL "bilan assurance CIC.PDF" "base64" 459532 NIL ("attachment" ("filename" "bilan assurance CIC.PDF" "size" "335811" "creation-date" "Fri, 16 Sep 2011 12:45:23 GMT" "modification-date" "Fri, 16 Sep 2011 12:45:23 GMT")) NIL NIL)(("text" "plain" ("charset" "utf-8") NIL NIL "quoted-printable" 1389 29 NIL NIL NIL NIL)("text" "html" ("charset" "utf-8") NIL NIL "quoted-printable" 1457 27 NIL NIL NIL NIL) "alternative" ("boundary" "===============0775904800==") ("inline" NIL) NIL NIL) "mixed" ("boundary" "_007_152AC7ECD1F8AB43A9AD95DBDDCA3118082C09GKIMA24cmcicfr_") NIL ("fr-FR") NIL)',
    ),
    b')'
]

BODYSTRUCTURE_SAMPLE_7 = [
    (b'856 (UID 11111 BODYSTRUCTURE ((("text" "plain" ("charset" "UTF-8") NIL NIL "7bit" 0 0 NIL NIL NIL NIL) "mixed" ("boundary" "----=_Part_407172_3159001.1321948277321") NIL NIL NIL)("application" "octet-stream" ("name" "26274308.pdf") NIL NIL "base64" 14906 NIL ("attachment" ("filename" "(26274308.pdf")) NIL NIL) "mixed" ("boundary" "----=_Part_407171_9686991.1321948277321") NIL NIL NIL)',)
    ,
    b')'
]

BODYSTRUCTURE_SAMPLE_8 = [
    (b'1 (UID 947 BODYSTRUCTURE ("text" "html" ("charset" "utf-8") NIL NIL "8bit" 889 34 NIL NIL NIL NIL) BODY[HEADER.FIELDS (FROM TO CC DATE SUBJECT)] {80}', 'From: Alphamonak Solutions <amonak@alphamonak.com>\r\nDate: Sat, 26 Mar 2016 11:45:49 +0100\r\n\r\n'),
    b')'
]

BODYSTRUCTURE_SAMPLE_9 = [
    (b'855 (UID 46932 ' + BODYSTRUCTURE_4 + b' BODY[HEADER.FIELDS (FROM TO CC DATE SUBJECT)] {235}', b'From: <Service.client10@maaf.fr>\r\nTo: <TONIO@NGYN.ORG>\r\nCc: \r\nSubject: Notre contact du 28/12/2011 - 192175092\r\nDate: Wed, 28 Dec 2011 13:29:17 +0100\r\nMessage-ID: <CABY0dkJspTaFn7v-1OG1nc9M0Qxn+VUTpcXzxyGNBnSnZtqMrw@mail.gmail.com>\r\n\r\n'),
    b')'
]

BODYSTRUCTURE_SAMPLE_10 = [
    (b'855 (UID 46932 ' + BODYSTRUCTURE_4 + b' BODY[1.1] {10})',
     b'XXXXXXXX\r\n',
     b'BODY[2] {10}',
     b'XXXXXXXX\r\n',
     b')',)
]

BODYSTRUCTURE_EMPTY_MAIL = [
    b'33 (UID 33 BODYSTRUCTURE (("text" "plain" ("charset" "us-ascii") NIL NIL "quoted-printable" 0 0 NIL NIL NIL NIL)("application" "zip" ("name" "hotmail.com!ngyn.org!1435428000!1435514400.zip") NIL NIL "base64" 978 NIL ("attachment" ("filename" "hotmail.com!ngyn.org!1435428000!1435514400.zip")) NIL NIL) "mixed" ("boundary" "--boundary_32281_b52cf564-2a50-4f96-aeb0-ef5f83b05463") NIL NIL NIL))'
]

EMPTY_BODY = [(b'33 (UID 33 BODY[1] {0}', b''), b')']

COMPLETE_MAIL = [
    (b'109 (UID 133872 BODY[] {2504}', b'Return-Path: <test1@test.org>\r\nDelivered-To: test1@test.org\r\nReceived: from mail.testsrv.org\r\n\tby nelson.ngyn.org (Dovecot) with LMTP id E9jPFPHHe1u1bQAABvoInA\r\n\tfor <test1@test.org>; Tue, 21 Aug 2018 10:06:09 +0200\r\nReceived: from [192.168.0.12] (unknown [109.9.172.169])\r\n\t(using TLSv1.2 with cipher ECDHE-RSA-AES256-GCM-SHA384 (256/256 bits))\r\n\t(No client certificate requested)\r\n\tby mail.testsrv.org (Postfix) with ESMTPSA id 2B279E00D2\r\n\tfor <test1@test.org>; Tue, 21 Aug 2018 10:06:09 +0200 (CEST)\r\nDKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed; d=test.org; s=test;\r\n\tt=1534838769; h=from:from:sender:reply-to:subject:subject:date:date:\r\n\t message-id:message-id:to:to:cc:mime-version:mime-version:\r\n\t content-type:content-type:\r\n\t content-transfer-encoding:content-transfer-encoding:in-reply-to:\r\n\t references; bh=EVfAHeUMDygbJe0SkMWJHjgXGjtiTLZnMQbyWqzsrCY=;\r\n\tb=fRc4i+WbPkbdD1BNfV9TqsZZ9nTbnHn6CKbH4nwrQnmUYQvvFUcpXC8PBURyglyLBAqPKb\r\n\tF6Dq7TnvcYdkIpR0XyMko+XB/qt8Wj0J0tC8bFxfknN8yNqj32SQAjtLrsjtzEgC1LkcLo\r\n\tcXU+FQbZR0D8EL7YD/nnsbO4mYnf3as=\r\nTo: Alphamonak Solutions <test1@test.org>\r\nFrom: Alphamonak Solutions <test1@test.org>\r\nSubject: Simple message\r\nMessage-ID: <7ed6e950-9e2b-b0fc-9e66-5b8358453b8d@test.org>\r\nDate: Tue, 21 Aug 2018 10:06:08 +0200\r\nUser-Agent: Mozilla/5.0 (X11; Linux x86_64; rv:52.0) Gecko/20100101\r\n Thunderbird/52.9.1\r\nMIME-Version: 1.0\r\nContent-Type: text/plain; charset=utf-8\r\nContent-Transfer-Encoding: 7bit\r\nContent-Language: fr-FR\r\nARC-Message-Signature: i=1; a=rsa-sha256; c=relaxed/relaxed; d=test.org;\r\n\ts=test; t=1534838769; h=from:from:sender:reply-to:subject:subject:date:date:\r\n\t message-id:message-id:to:to:cc:mime-version:mime-version:\r\n\t content-type:content-type:\r\n\t content-transfer-encoding:content-transfer-encoding:in-reply-to:\r\n\t references; bh=EVfAHeUMDygbJe0SkMWJHjgXGjtiTLZnMQbyWqzsrCY=;\r\n\tb=ZV1ndumHftEnhq8Z7zujDS9WUlpJmjrTIEzXsIy8+1HhZ6AEcmur5r8SmS/xfUvpoScNvF\r\n\tlYAHIdbKPX54Fpq7rZVgIhuy8vfcRGprHxH9DMhk3kHiuZLsJT1F6UdPwTYSGAWBTbg1u2\r\n\tHu2VoeSYBV5PrsINvPr1QCbo+GqTlJU=\r\nARC-Seal: i=1; s=test; d=test.org; t=1534838769; a=rsa-sha256; cv=none;\r\n\tb=CE7Q0SnkLMHx82eOtanv4pCtl3fhaPqOAYb3Nqxi5lqwDcO80vTMfGUwOv6r3Exv9Cnj3+nCxejzXtrO4TTKesL9bamOBf+veRYTLfy8wjRIfgEVXS/amOjf9u4UzGgwxL1HbhHWbCIqJhaVwrPZaHZwA9XPmboqGVQNP3i2nH0=\r\nARC-Authentication-Results: i=1; ORIGINATING;\r\n\tauth=pass smtp.auth=test1@test.org smtp.mailfrom=test1@test.org\r\nAuthentication-Results: ORIGINATING;\r\n\tauth=pass smtp.auth=test1@test.org smtp.mailfrom=test1@test.org\r\n\r\nHello!\r\n\r\n'), b')'
]



# -*- coding: utf-8 -*-

"""DMARC related tools for Kalabash."""

from __future__ import unicode_literals

from pkg_resources import get_distribution, DistributionNotFound


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = '9.9.9'

default_app_config = "kalabash_webmail.apps.WebmailConfig"



from email import encoders
from email.mime.base import MIMEBase
import os
import uuid

import six

from django.conf import settings
from django.core.files.uploadhandler import FileUploadHandler, SkipFile
from django.utils.encoding import smart_bytes

from kalabash.lib.exceptions import InternalError
from kalabash.lib.web_utils import size2integer
from kalabash.parameters import tools as param_tools

from .rfc6266 import build_header


def get_storage_path(filename):
    """Return a path to store a file."""
    storage_dir = os.path.join(settings.MEDIA_ROOT, "webmail")
    if not filename:
        return storage_dir
    return os.path.join(storage_dir, filename)


def set_compose_session(request):
    """Initialize a new "compose" session.

    It is used to keep track of attachments defined with a new
    message. Each new message will be associated with a unique ID (in
    order to avoid conflicts between users).

    :param request: a Request object.
    :return: the new unique ID.
    """
    randid = str(uuid.uuid4()).replace("-", "")
    request.session["compose_mail"] = {"id": randid, "attachments": []}
    return randid


def save_attachment(f):
    """Save a new attachment to the filesystem.

    The attachment is not saved using its own name to the
    filesystem. To avoid conflicts, a random name is generated and
    used instead.

    :param f: an uploaded file object (see Django's documentation) or bytes
    :return: the new random name
    """
    from tempfile import NamedTemporaryFile

    try:
        fp = NamedTemporaryFile(dir=get_storage_path(""), delete=False)
    except Exception as e:
        raise InternalError(str(e))
    if isinstance(f, (six.binary_type, six.text_type)):
        fp.write(smart_bytes(f))
    else:
        for chunk in f.chunks():
            fp.write(chunk)
    fp.close()
    return fp.name


def clean_attachments(attlist):
    """Remove all attachments from the filesystem

    :param attlist: a list of 2-uple. Each element must contain the
                    following information : (random name, real name).
    """
    for att in attlist:
        fullpath = get_storage_path(att["tmpname"])
        try:
            os.remove(fullpath)
        except OSError:
            pass


def create_mail_attachment(attdef, payload=None):
    """Create the MIME part corresponding to the given attachment.

    Mandatory keys: 'fname', 'tmpname', 'content-type'

    :param attdef: a dictionary containing the attachment definition
    :return: a MIMEBase object
    """
    if "content-type" in attdef:
        maintype, subtype = attdef["content-type"].split("/")
    elif "Content-Type" in attdef:
        maintype, subtype = attdef["Content-Type"].split("/")
    else:
        return None
    res = MIMEBase(maintype, subtype)
    if payload is None:
        path = get_storage_path(attdef["tmpname"])
        with open(path, "rb") as fp:
            res.set_payload(fp.read())
    else:
        res.set_payload(payload)
    encoders.encode_base64(res)
    if isinstance(attdef["fname"], six.binary_type):
        attdef["fname"] = attdef["fname"].decode("utf-8")
    content_disposition = build_header(attdef["fname"])
    if isinstance(content_disposition, six.binary_type):
        res["Content-Disposition"] = content_disposition.decode("utf-8")
    else:
        res["Content-Disposition"] = content_disposition
    return res


class AttachmentUploadHandler(FileUploadHandler):

    """
    Simple upload handler to limit the size of the attachments users
    can upload.
    """

    def __init__(self, request=None):
        super(AttachmentUploadHandler, self).__init__(request)
        self.total_upload = 0
        self.toobig = False
        self.maxsize = size2integer(
            param_tools.get_global_parameter("max_attachment_size"))

    def receive_data_chunk(self, raw_data, start):
        self.total_upload += len(raw_data)
        if self.total_upload >= self.maxsize:
            self.toobig = True
            raise SkipFile()
        return raw_data

    def file_complete(self, file_size):
        return None



# coding: utf-8
"""
:mod:`exceptions` --- Webmail custom exceptions
-----------------------------------------------

"""
import re

from django.utils.translation import gettext as _

from kalabash.lib.exceptions import KalabashException, InternalError


class WebmailInternalError(InternalError):
    errorexpr = re.compile(r'\[([^\]]+)\]\s*([^\.]+)')

    def __init__(self, reason, ajax=False):
        match = WebmailInternalError.errorexpr.match(reason)
        if not match:
            self.reason = reason
        else:
            self.reason = "%s: %s" % (_("Server response"), match.group(2))
        self.ajax = ajax

    def __str__(self):
        return self.reason


class UnknownAction(KalabashException):

    """
    Use this exception when the webmail encounter an unknown action.
    """
    http_code = 404

    def __init__(self):
        super(UnknownAction, self).__init__(_("Unknown action"))


class ImapError(KalabashException):

    http_code = 500

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return str(self.reason)



# coding: utf-8

"""Webmail forms."""

from email.header import Header
from email.mime.image import MIMEImage
import os
import pkg_resources

from six.moves.urllib.parse import urlparse, unquote

import lxml.html

from django import forms
from django.conf import settings
from django.core.mail import EmailMessage, EmailMultiAlternatives
from django.core.validators import validate_email
from django.utils.translation import gettext_lazy as _

from ckeditor_uploader.widgets import CKEditorUploadingWidget

from kalabash.lib import email_utils, form_utils
from kalabash.parameters import forms as param_forms

from .lib import (
    ImapEmail, create_mail_attachment, decode_payload
)
from .validators import validate_email_list


def html2plaintext(content):
    """HTML to plain text translation.

    :param content: some HTML content
    """
    if not content:
        return ""
    html = lxml.html.fromstring(content)
    plaintext = ""
    for ch in html.iter():
        p = None
        if ch.text is not None:
            p = ch.text.strip('\r\t\n')
        if ch.tag == "img":
            p = ch.get("alt")
        if p is None:
            continue
        plaintext += p + "\n"
    return plaintext


def make_body_images_inline(body):
    """Looks for images inside the body and make them inline.

    Before sending a message in HTML format, it is necessary to find
    all img tags contained in the body in order to rewrite them. For
    example, icons provided by CKeditor are stored on the server
    filesystem and not accessible from the outside. We must embark
    them as parts of the MIME message if we want recipients to
    display them correctly.

    :param body: the HTML body to parse
    """
    html = lxml.html.fromstring(body)
    parts = []
    for tag in html.iter("img"):
        src = tag.get("src")
        if src is None:
            continue
        o = urlparse(src)
        path = unquote(os.path.join(settings.BASE_DIR, o.path[1:]))
        if not os.path.exists(path):
            continue
        fname = os.path.basename(path)
        cid = "%s@kalabash" % os.path.splitext(fname)[0]
        tag.set("src", "cid:%s" % cid)
        with open(path, "rb") as fp:
            part = MIMEImage(fp.read())
        part["Content-ID"] = "<%s>" % cid
        part.replace_header(
            "Content-Type", '%s; name="%s"' % (part["Content-Type"], fname)
        )
        part["Content-Disposition"] = "inline"
        parts.append(part)
    return lxml.html.tostring(html, encoding="unicode"), parts


class ComposeMailForm(forms.Form):
    """Compose mail form."""

    from_ = forms.ChoiceField(
        label=_("From"), choices=[],
        widget=forms.Select(attrs={"class": "selectize"})
    )
    to = forms.CharField(
        label=_("To"), validators=[validate_email_list])
    cc = forms.CharField(
        label=_("Cc"), required=False, validators=[validate_email_list],
        widget=forms.TextInput(
            attrs={"placeholder": _("Enter one or more addresses."),
                   "class": "selectize-contact"})
    )
    bcc = forms.CharField(
        label=_("Bcc"), required=False, validators=[validate_email_list],
        widget=forms.TextInput(
            attrs={"placeholder": _("Enter one or more addresses."),
                   "class": "selectize-contact"})
    )

    subject = forms.CharField(
        label=_("Subject"),
        max_length=255,
        required=False
    )
    origmsgid = forms.CharField(
        label="", required=False, widget=forms.HiddenInput())
    body = forms.CharField(
        required=False, widget=CKEditorUploadingWidget(
            attrs={"class": "editor form-control"})
    )

    def __init__(self, user, *args, **kwargs):
        """Custom constructor."""
        super(ComposeMailForm, self).__init__(*args, **kwargs)
        from_addresses = [(user.email, user.email)]
        for address in user.mailbox.alias_addresses:
            try:
                validate_email(address)
                from_addresses += [(address, address)]
            except forms.ValidationError:
                pass
        additional_sender_addresses = (
            user.mailbox.senderaddress_set.values_list("address", flat=True))
        from_addresses += [
            (address, address) for address in additional_sender_addresses]
        self.fields["from_"].choices = from_addresses
        self.fields["from_"].initial = user.email
        self.field_widths = {
            "cc": "11",
            "bcc": "11",
            "subject": "11"
        }

    def clean_to(self):
        """Convert to a list."""
        to = self.cleaned_data["to"]
        return email_utils.prepare_addresses(to, "envelope")

    def clean_cc(self):
        """Convert to a list."""
        cc = self.cleaned_data["cc"]
        return email_utils.prepare_addresses(cc, "envelope")

    def clean_bcc(self):
        """Convert to a list."""
        bcc = self.cleaned_data["bcc"]
        return email_utils.prepare_addresses(bcc, "envelope")

    def _html_msg(self, sender, headers):
        """Create a multipart message.

        We attach two alternatives:
        * text/html
        * text/plain
        """
        body = self.cleaned_data["body"]
        if body:
            tbody = html2plaintext(body)
            body, images = make_body_images_inline(body)
        else:
            tbody = ""
            images = []
        msg = EmailMultiAlternatives(
            self.cleaned_data["subject"],
            tbody,
            sender, self.cleaned_data["to"],
            cc=self.cleaned_data["cc"],
            bcc=self.cleaned_data["bcc"],
            headers=headers
        )
        msg.attach_alternative(body, "text/html")
        for img in images:
            msg.attach(img)
        return msg

    def _plain_msg(self, sender, headers):
        """Create a simple text message."""
        msg = EmailMessage(
            self.cleaned_data["subject"],
            self.cleaned_data["body"],
            sender,
            self.cleaned_data["to"],
            cc=self.cleaned_data["cc"],
            bcc=self.cleaned_data["bcc"],
            headers=headers
        )
        return msg

    def _format_sender_address(self, user, address):
        """Format address before message is sent."""
        if user.first_name != "" or user.last_name != "":
            return '"{}" <{}>'.format(
                Header(user.fullname, "utf8"), address)
        return address

    def _build_msg(self, request):
        """Build message to send.

        Can be overidden by children.
        """
        headers = {
            "User-Agent": "Kalabash {}".format(
                pkg_resources.get_distribution("kalabash").version)
        }
        origmsgid = self.cleaned_data.get("origmsgid")
        if origmsgid:
            headers.update({
                "References": origmsgid,
                "In-Reply-To": origmsgid
            })
        mode = request.user.parameters.get_value("editor")
        sender = self._format_sender_address(
            request.user, self.cleaned_data["from_"])
        return getattr(self, "_{}_msg".format(mode))(sender, headers)

    def to_msg(self, request):
        """Convert form's content to an object ready to send."""
        msg = self._build_msg(request)
        if request.session["compose_mail"]["attachments"]:
            for attdef in request.session["compose_mail"]["attachments"]:
                msg.attach(create_mail_attachment(attdef))
        return msg


class ForwardMailForm(ComposeMailForm):
    """Forward mail form."""

    def _build_msg(self, request):
        """Convert form's content to a MIME message.

        We also add original attachments (if any) to the new message.
        """
        mbox = request.GET.get("mbox", None)
        mailid = request.GET.get("mailid", None)
        msg = super(ForwardMailForm, self)._build_msg(request)
        origmsg = ImapEmail(request, "%s:%s" % (mbox, mailid))
        if origmsg.attachments:
            for attpart, fname in origmsg.attachments.items():
                attdef, payload = origmsg.fetch_attachment(attpart)
                attdef["fname"] = fname
                msg.attach(create_mail_attachment(
                    attdef, decode_payload(attdef["encoding"], payload)
                ))
        return msg


class FolderForm(forms.Form):
    oldname = forms.CharField(
        label="",
        widget=forms.HiddenInput(attrs={"class": "form-control"}),
        required=False
    )
    name = forms.CharField(
        widget=forms.TextInput(attrs={"class": "form-control"}))


class AttachmentForm(forms.Form):
    attachment = forms.FileField(label=_("Select a file"), allow_empty_file=True)


class ParametersForm(param_forms.AdminParametersForm):
    app = "kalabash_webmail"

    sep3 = form_utils.SeparatorField(label=_("General"))

    max_attachment_size = forms.CharField(
        label=_("Maximum attachment size"),
        initial="2048",
        help_text=_(
            "Maximum attachment size in bytes (or KB, MB, GB if specified)")
    )

    sep1 = form_utils.SeparatorField(label=_("IMAP settings"))

    imap_server = forms.CharField(
        label=_("Server address"),
        initial="127.0.0.1",
        help_text=_("Address of your IMAP server")
    )

    imap_secured = form_utils.YesNoField(
        label=_("Use a secured connection"),
        initial=False,
        help_text=_("Use a secured connection to access IMAP server")
    )

    imap_port = forms.IntegerField(
        label=_("Server port"),
        initial=143,
        help_text=_("Listening port of your IMAP server")
    )

    sep2 = form_utils.SeparatorField(label=_("SMTP settings"))

    smtp_server = forms.CharField(
        label=_("Server address"),
        initial="127.0.0.1",
        help_text=_("Address of your SMTP server")
    )

    smtp_secured_mode = forms.ChoiceField(
        label=_("Secured connection mode"),
        choices=[("none", _("None")),
                 ("starttls", "STARTTLS"),
                 ("ssl", "SSL/TLS")],
        initial="none",
        help_text=_("Use a secured connection to access SMTP server"),
        widget=form_utils.HorizontalRadioSelect()
    )

    smtp_port = forms.IntegerField(
        label=_("Server port"),
        initial=25,
        help_text=_("Listening port of your SMTP server")
    )

    smtp_authentication = form_utils.YesNoField(
        label=_("Authentication required"),
        initial=False,
        help_text=_("Server needs authentication")
    )


class UserSettings(param_forms.UserParametersForm):
    app = "kalabash_webmail"

    sep1 = form_utils.SeparatorField(label=_("Display"))

    displaymode = forms.ChoiceField(
        initial="plain",
        label=_("Default message display mode"),
        choices=[("html", "html"), ("plain", "text")],
        help_text=_("The default mode used when displaying a message"),
        widget=form_utils.HorizontalRadioSelect()
    )

    enable_links = form_utils.YesNoField(
        initial=False,
        label=_("Enable HTML links display"),
        help_text=_("Enable/Disable HTML links display")
    )

    messages_per_page = forms.IntegerField(
        initial=40,
        label=_("Number of displayed emails per page"),
        help_text=_("Sets the maximum number of messages displayed in a page")
    )

    refresh_interval = forms.IntegerField(
        initial=300,
        label=_("Listing refresh rate"),
        help_text=_("Automatic folder refresh rate (in seconds)")
    )

    mboxes_col_width = forms.IntegerField(
        initial=200,
        label=_("Folder container's width"),
        help_text=_("The width of the folder list container")
    )

    sep2 = form_utils.SeparatorField(label=_("Folders"))

    trash_folder = forms.CharField(
        initial="Trash",
        label=_("Trash folder"),
        help_text=_("Folder where deleted messages go")
    )

    sent_folder = forms.CharField(
        initial="Sent",
        label=_("Sent folder"),
        help_text=_("Folder where copies of sent messages go")
    )

    drafts_folder = forms.CharField(
        initial="Drafts",
        label=_("Drafts folder"),
        help_text=_("Folder where drafts go")
    )
    junk_folder = forms.CharField(
        initial="Junk",
        label=_("Junk folder"),
        help_text=_("Folder where junk messages should go")
    )

    sep3 = form_utils.SeparatorField(label=_("Composing messages"))

    editor = forms.ChoiceField(
        initial="plain",
        label=_("Default editor"),
        choices=[("html", "html"), ("plain", "text")],
        help_text=_("The default editor to use when composing a message"),
        widget=form_utils.HorizontalRadioSelect()
    )

    signature = forms.CharField(
        initial="",
        label=_("Signature text"),
        help_text=_("User defined email signature"),
        required=False,
        widget=CKEditorUploadingWidget()
    )

    visibility_rules = {
        "enable_links": "displaymode=html"
    }

    @staticmethod
    def has_access(**kwargs):
        return hasattr(kwargs.get("user"), "mailbox")

    def clean_mboxes_col_width(self):
        """Check if the entered value is a positive integer.

        It must also be different from 0.
        """
        if self.cleaned_data['mboxes_col_width'] <= 0:
            raise forms.ValidationError(
                _('Value must be a positive integer (> 0)')
            )
        return self.cleaned_data['mboxes_col_width']



"""
Django settings for test_project project.

Generated by 'django-admin startproject' using Django 1.11.8.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

from logging.handlers import SysLogHandler
import os

from kalabash.test_settings import *  # noqa


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'w537@nm@5n)=+e%-7*z-jxf21a#0k%uv^rbu**+cj4=_u57e(8'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = 'DEBUG' in os.environ

TEMPLATE_DEBUG = DEBUG

ALLOWED_HOSTS = [
    '127.0.0.1',
    'localhost',
]

SITE_ID = 1

# Security settings

X_FRAME_OPTIONS = "SAMEORIGIN"
CSRF_COOKIE_SECURE = False
SESSION_COOKIE_SECURE = False

# Application definition

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'reversion',
    'ckeditor',
    'ckeditor_uploader',
    'rest_framework',
    'rest_framework.authtoken',
    'drf_spectacular',
    'django_otp',
    'django_otp.plugins.otp_totp',
    'django_otp.plugins.otp_static',
    'django_rename_app',
)

# A dedicated place to register Kalabash applications
# Do not delete it.
# Do not change the order.
KALABASH_APPS = (
    'kalabash',
    'kalabash.core',
    'kalabash.lib',
    'kalabash.admin',
    'kalabash.transport',
    'kalabash.relaydomains',
    'kalabash.limits',
    'kalabash.parameters',
    'kalabash.dnstools',
    'kalabash.policyd',
    'kalabash.maillog',
    'kalabash.pdfcredentials',
    'kalabash.dmarc',
    'kalabash.imap_migration',
    # Kalabash extensions here.
    'kalabash_webmail',
)

INSTALLED_APPS += KALABASH_APPS

AUTH_USER_MODEL = 'core.User'

MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'x_forwarded_for.middleware.XForwardedForMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django_otp.middleware.OTPMiddleware',
    'kalabash.core.middleware.TwoFAMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'kalabash.core.middleware.LocalConfigMiddleware',
    'kalabash.lib.middleware.AjaxLoginRedirect',
    'kalabash.lib.middleware.CommonExceptionCatcher',
    'kalabash.lib.middleware.RequestCatcherMiddleware',
)

AUTHENTICATION_BACKENDS = (
    # 'kalabash.lib.authbackends.LDAPBackend',
    # 'kalabash.lib.authbackends.SMTPBackend',
    'django.contrib.auth.backends.ModelBackend',
    'kalabash.imap_migration.auth_backends.IMAPBackend',
)

# SMTP authentication
# AUTH_SMTP_SERVER_ADDRESS = 'localhost'
# AUTH_SMTP_SERVER_PORT = 25
# AUTH_SMTP_SECURED_MODE = None  # 'ssl' or 'starttls' are accepted


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
                'kalabash.core.context_processors.top_notifications',
            ],
            'debug': TEMPLATE_DEBUG,
        },
    },
]

ROOT_URLCONF = 'test_project.urls'

WSGI_APPLICATION = 'test_project.wsgi.application'


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

STATIC_URL = '/sitestatic/'
STATIC_ROOT = os.path.join(BASE_DIR, 'sitestatic')
STATICFILES_DIRS = (
    os.path.join(BASE_DIR, '..', 'kalabash', 'bower_components'),
)

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

# Rest framework settings

REST_FRAMEWORK = {
    'DEFAULT_THROTTLE_RATES': {
        'user': '1000/minute',
        'ddos': '200/second',
        'ddos_lesser': '1000/minute',
        'login': '100/minute',
        'password_recovery_request': '11/hour',
        'password_recovery_totp_check': '25/hour',
        'password_recovery_apply': '25/hour'
    },
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'kalabash.core.drf_authentication.JWTAuthenticationWith2FA',
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
    'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.NamespaceVersioning',
}

# Kalabash settings
# KALABASH_CUSTOM_LOGO = os.path.join(MEDIA_URL, "custom_logo.png")

# DOVECOT_LOOKUP_PATH = ('/path/to/dovecot', )

KALABASH_API_URL = 'https://api.alphamonak.com/1/'

# REDIS

REDIS_HOST = os.environ.get('REDIS_HOST', '127.0.0.1')
REDIS_PORT = os.environ.get('REDIS_PORT', 6379)
REDIS_QUOTA_DB = 0
REDIS_URL = 'redis://{}:{}/{}'.format(REDIS_HOST, REDIS_PORT, REDIS_QUOTA_DB)

# RQ

RQ_QUEUES = {
    'dkim': {
        'URL': REDIS_URL,
    },
    'kalabash': {
        'URL': REDIS_URL,
    },
}

# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
    {
        'NAME': 'kalabash.core.password_validation.ComplexityValidator',
        'OPTIONS': {
            'upper': 1,
            'lower': 1,
            'digits': 1,
            'specials': 0
        }
    },
]

# CKeditor

CKEDITOR_UPLOAD_PATH = "uploads/"

CKEDITOR_IMAGE_BACKEND = "pillow"

CKEDITOR_RESTRICT_BY_USER = True

CKEDITOR_BROWSE_SHOW_DIRS = True

CKEDITOR_ALLOW_NONIMAGE_FILES = False

CKEDITOR_CONFIGS = {
    'default': {
        'allowedContent': True,
        'toolbar': 'Kalabash',
        'width': None,
        'toolbar_Kalabash': [
            ['Bold', 'Italic', 'Underline'],
            ['JustifyLeft', 'JustifyCenter', 'JustifyRight', 'JustifyBlock'],
            ['BidiLtr', 'BidiRtl', 'Language'],
            ['NumberedList', 'BulletedList', '-', 'Outdent', 'Indent'],
            ['Undo', 'Redo'],
            ['Link', 'Unlink', 'Anchor', '-', 'Smiley'],
            ['TextColor', 'BGColor', '-', 'Source'],
            ['Font', 'FontSize'],
            ['Image', ],
            ['SpellChecker']
        ],
    },
}

# Logging configuration

LOGGING = {
    'version': 1,
    'formatters': {
        'syslog': {
            'format': '%(name)s: %(levelname)s %(message)s'
        },
    },
    'handlers': {
        'syslog-auth': {
            'class': 'logging.handlers.SysLogHandler',
            'facility': SysLogHandler.LOG_AUTH,
            'formatter': 'syslog'
        },
        'kalabash': {
            'class': 'kalabash.core.loggers.SQLHandler',
        }
    },
    'loggers': {
        'kalabash.auth': {
            'handlers': ['syslog-auth', 'kalabash'],
            'level': 'INFO',
            'propagate': False
        },
        'kalabash.admin': {
            'handlers': ['kalabash'],
            'level': 'INFO',
            'propagate': False
        }
    }
}

# Load settings from extensions
